//
//  BallTree.swift
//  Oritatami Builder
//
//  Created by Nikao on 18/09/2019.
//  Copyright © 2019 Nikao. All rights reserved.
//

import UIKit

//struct PathTree : CustomStringConvertible{
//    var child : [Direction:PathTree]
//    var nbBonds : Int
//
//
//    init(_ dict: [Direction: PathTree] = [:], nbBonds nb: Int = 0) {
//        child = dict
//        nbBonds = nb
//    }
//
//    mutating func insert(path: Path, i: Int = 0) {
//        if i >= path.count {
//            print("Path already in the path Tree !")
//        } else {
//            let d = path[i]
//            if var a = child[d] {
//                a.insert(path: path, i: i+1)
//            } else {
//                print("insert \(d)")
//                child[d] = PathTree([:])
//                if i < path.count - 1 {
//                   child[d]!.insert(path: path, i: i+1)
//                }
//            }
//        }
//    }
//
//    func description(offset: String = "") -> String {
//        guard child.count > 0 else { return "\n" } // leaf
//
//        var s = ""
//        print(child)
//        for d in Direction.all {
//            if let a = child[d] {
//                print("print \(d)")
//                s += offset + d.description + "\n" + a.description(offset: offset + "  ")
//            }
//        }
//        return s
//    }
//
//    var description: String {
//        return description()
//    }
//}

// ONLY FOR DELAY >=2

class PathBall : CustomStringConvertible {
    var paths = [Path]()
    var commonPrefixLength = 0
    var transcriptIndex = 0 // 1st bead of the path (at the end of path[0][0])
    var nbBonds = 0
    var origin = IntPoint.zero
    
    var count : Int { return paths.count }
    
    var direction : Direction? {
        return paths.count == 0 ? nil : paths[0][0]
    }
    
    required init(origin o: IntPoint = IntPoint.zero) {
        origin = o
    }
    
    func insert(path: Path) {
        if paths.count == 0 {
            paths.append(path)
            commonPrefixLength = path.count
            return
        }

        for i in 0..<commonPrefixLength {
            if path[i] != paths[0][i] {
                commonPrefixLength = i
                break
            }
        }
        if commonPrefixLength == 0 {
            print("ERROR: path with incorrect first step \(path[0]) inserted in path ball with first step \(paths[0][0])")
        }
        
        paths.append(path)
    }
    
    func split() -> [PathBall] {
        var newBalls = [PathBall?](repeating: nil, count: Direction.count)
        for p in paths {
            let d = p[1]
            if newBalls[d.rawValue] == nil {
                newBalls[d.rawValue] = PathBall()
            }
            newBalls[d.rawValue]!.insert(path: Array(p[1..<p.count]))
        }
        return newBalls.filter() { $0 != nil } as! [PathBall]
    }
    
    var description: String {
        guard paths.count > 0 else { return "[empty]" }

        var s = paths[0][0..<commonPrefixLength].description + "<" + paths[0][commonPrefixLength..<paths[0].count].description
        for i in 1..<paths.count {
            s += ", " + paths[i][commonPrefixLength..<paths[i].count].description
        }
        s += ">"
        return s
    }
}


class EmbeddedPathBall : PathBall {
    var center = CGPoint.zero // englobing circle center in the triangular grid
    var radius : CGFloat = 0 // englobing circle radius
    
    var position = CGPoint.zero // center position on the PDF page
    
    func updateCircle() {
        (center, radius) = CGPoint.englobingCircle(of: paths, from: origin)
        radius += Defaults.englobingCircleExtraRadius
    }
    
    var northAnchor : CGPoint { return position + CGPoint(x: 0, y:  -radius) }
    var southAnchor : CGPoint { return position + CGPoint(x: 0, y: radius) }
    var rect : CGRect { return position + radius * CGRect(x: -1, y: -1, width: 2, height: 2) }
    var drawRect : CGRect { return center + radius * CGRect(x: -1, y: -1, width: 2, height: 2) }
    
    func draw(in ctx: __CGContext, view: TriangularGridView, oritatamiSystem os: OritatamiSystem) {
        print("Draw Ball")
        guard paths.count > 0 else { return }
        guard let grid = view.grid else { return }
        
        let currentCtx = __GraphicsGetCurrentContext()
        if ctx != currentCtx {
            __GraphicsPushContext(ctx)
        }
        
        let savedScale = view.scale
        view.scale = TriangularGridView.Defaults.scale
        
        // Add the common part of the path in the grid
        var q = origin
        let path0 = paths[0]
        for i in 0..<commonPrefixLength {
            let d = path0[i]
            let v = q.neighbor(d)
            let j = transcriptIndex + i
            let vB = os.beadTypeId(indexed: j)
            grid.synchronize {
                let _ = grid.addBead(typeId: vB, from: q, to: d, isTest:true, isSeed: j < os.seedPath.count)
                grid.setBeadIndex(at: v, value: j)
                //                let _ = grid.addBead(typeId: os.beadTypeId(indexed:j), at: v, bond: .testBond)
                //                grid.setBeadIndex(at: v, value: j)
                //                grid.setOutEdge(at: q, to: d)
                //                grid.setBond(at: q, to: d,
                //                             value: j < os.seedPath.count ? .testSeed : .testPrimary)
            }
            q = v;
        }
        let endOfCommon = q
        
        // Draw the ball
        ctx.saveGState()
        
        let dRect = drawRect
        let ballRect = rect
        ctx.translateBy(x: position.x - center.x, y: position.y - center.y)
        ctx.addEllipse(in: center + radius * CGRect.unitCircle)
        ctx.clip()
        
        // Draw the grid
        view.drawBeads(grid, rect: center + radius * CGRect.unitCircle, in: ctx, drawGrid: true, drawBeads: false)
        
        // Draw the path
        let pathLength = path0.count
        if commonPrefixLength < pathLength {
            ctx.setLineWidth(Defaults.looseTestBondLineWidth)
            for k in 0..<paths.count {
                var offset = CGPoint.zero
                var radius = Defaults.looseTestBeadLargeRadius
                if commonPrefixLength < pathLength-1 {
                    offset = Defaults.looseTestBeadOffset *
                        CGPoint(x: 2 * (0.5 - drand48()), y: 2 * (0.5 - drand48()))
                    radius = Defaults.looseTestBeadRadius
                }
                
                let h = CGFloat(k)/CGFloat(paths.count)
                let br = CGFloat(1+k%2)/2
                let c = __Color(hue: h, saturation: 1, brightness:br, alpha:1)
                ctx.setStrokeColor(c.cgColor)
                ctx.setFillColor(c.withAlphaComponent(0.2).cgColor)
                
                var q = endOfCommon
                let beadRect = radius * CGRect.unitCircle
                let path = paths[k]
                for i in commonPrefixLength..<pathLength {
                    let v = q.neighbor(path[i])
                    let gq = q.point + offset
                    let gv = v.point + offset
                    let a = atan2(gv.y - gq.y, gv.x - gq.x)
                    let Δ = radius * CGPoint(x: cos(a), y: sin(a))
                    let gq1 = gq + Δ
                    let gv1 = gv - Δ
                    ctx.move(to: gq1)
                    ctx.addLine(to: gv1)
                    ctx.addEllipse(in: gv + beadRect)
                    ctx.drawPath(using: .fillStroke)
                    q = v
                }
            }
        }
        
        // Draw the beads
        view.drawBeads(grid, rect: center + radius * CGRect.unitCircle, in: ctx, drawGrid: false, drawBeads: true)
        
        ctx.restoreGState()
        
        // Erase path
        q = origin // neighborOf(self.pathOrigin, oppositeDirection(decodeIthMove(self.pathCode, pathLength, 0)));
        for i in 0..<commonPrefixLength {
            let v = q.neighbor(path0[i])
            grid.removeBead(at: v)
            q = v
        }
        
        view.scale = savedScale
        
        if ctx != currentCtx {
            __GraphicsPopContext()
        }
    }
}

class EmbeddedPathBallTree: PathBallTree<EmbeddedPathBall> {
    var minChildX = CGFloat.infinity
    var maxChildX = -CGFloat.infinity
    var segmentChildY = CGFloat(0)
    
    var minFavorableChildX = CGFloat.infinity
    var maxFavorableChildX = -CGFloat.infinity

    func computeChildMinXMaxX() {
        minChildX = CGFloat.infinity
        maxChildX = -CGFloat.infinity
        
        minFavorableChildX = CGFloat.infinity
        maxFavorableChildX = -CGFloat.infinity
        
        for b in children {
            let x = b.pathBall.northAnchor.x
            if x < minChildX { minChildX = x }
            if x > maxChildX { maxChildX = x }
            if b.isFavorable {
                if x < minFavorableChildX { minFavorableChildX = x }
                if x > maxFavorableChildX { maxFavorableChildX = x }
            }
        }
    }
}

class PathBallTree<_PathBall: PathBall> {
    var pathBall: _PathBall
    weak var parent: PathBallTree? = nil { didSet { if let p = parent { p.add(child: self) } } }
    var children: [PathBallTree] = []
    
    var height : Int = 0
    var hasSelectedChild = false
    var hasFavorableChild = false
    var isSelected = false { didSet { if let p = parent, isSelected { p.hasSelectedChild = true } } } // is selected for the next step, i.e. has the same first step as the favorable configurations
    var isFavorable = false { didSet { if let p = parent, isFavorable { p.hasFavorableChild = true } } } // has maximum number of bonds
    var isDeadEnd = false
    
    init(pathBall b: _PathBall? = nil, parent p: PathBallTree<_PathBall>? = nil) {
        pathBall = b ?? _PathBall()
        parent = p
        if let p = parent { p.add(child: self) }
    }
    
    func add(child: PathBallTree<_PathBall>) {
        print("Child \(child) added to \(self)")
        child.parent = self
        child.height = height + 1
        children.append(child)
    }
    
    func removeChildren() {
        for c in children {
            c.removeChildren()
            c.parent = nil
        }
        children.removeAll()
    }
    

}

class PathBallLevelTree<_PathBall: PathBall>  {
    var levels = [[PathBallTree<_PathBall>]]()
    var transcriptPath = Path()
    var currentLevel = -1
    
    var yLevels = [CGFloat]()
    var hLevels = [CGFloat]()

    var environments = [Environment]()
    
    func moveLowestSelectedLevelToTop() {
        let last = levels.last!.filter() { $0.isSelected && !$0.isDeadEnd }
        levels.removeAll()
        levels.append(last)
        for t in last {
            t.parent = nil
        }
        currentLevel = 0
        transcriptPath = [levels[0][0].pathBall.paths[0][0]]
    }
}

class EmbeddedPathBallLevelTree : PathBallLevelTree<EmbeddedPathBall>  {
    
    func placeBalls(in width: CGFloat) {
        yLevels = [CGFloat](repeating: 0, count: levels.count)
        hLevels = [CGFloat](repeating: 0, count: levels.count)

        // From Bottom to Top
        
        var y = Defaults.ProofTreePDF.pageMarginY - Defaults.ProofTreePDF.interlevelMargin
        
        for l in stride(from: levels.count-1, through: 0, by: -1) {
            yLevels[l] = y + Defaults.ProofTreePDF.interLayerMargin
            
            let lev = levels[l]

            guard lev.count > 0 else { print("Empty level ?!?"); continue }
            
            // Compute the total width of the balls
            let sumRadius = lev.reduce(CGFloat(0)) { $0 + $1.pathBall.radius }
            let totalWidth = 2 * sumRadius + CGFloat(lev.count-1) * Defaults.ProofTreePDF.interchildMarginX
            
            // Compute the number of layers in this level
            let nbLayers = Int(ceil(min(totalWidth/width,
                                        1 + totalWidth/CGFloat(lev.count)/Defaults.ProofTreePDF.interchildMarginX)))
                
            var xLayer = [CGFloat](repeating: 0, count: nbLayers)
            var hLayer = [CGFloat](repeating: 0, count: nbLayers)
            var xLast : CGFloat = 0
            var minX : CGFloat = 0
            var xLastChild : CGFloat = 0
            var xLastParent : CGFloat = 0

            var bestLayer = 0
            var layer = [Int](repeating: 0, count: lev.count)
            var hSegment = 0
            var minHSegment = 0
            var maxHSegment = 0
            
            // Init X
            xLast = Defaults.ProofTreePDF.pageMarginX - Defaults.ProofTreePDF.interchildMarginX
            for i in 0..<nbLayers {
                xLayer[i] = xLast
                hLayer[i] = 0
            }
            
            // Place the pathBalls into the layers and set the segment accordingly
            
            hSegment = 0
            minHSegment = hSegment
            maxHSegment = hSegment
            xLastChild = -CGFloat.infinity
            xLastParent = -CGFloat.infinity
            
            for i in 0..<lev.count {
                let t = lev[i]
                let b = t.pathBall
                
                // Place on the leftmost layer
                minX = CGFloat.infinity
                bestLayer = -1
                let xMin = xLast - b.radius
                for j in 0..<nbLayers {
                    let x = max(xMin, xLayer[j])
                    if x < minX {
                        minX = x
                        bestLayer = j
                    }
                }
                minX += Defaults.ProofTreePDF.interchildMarginX
                
                // Place the ball in the leftmost layer and update it dans cette Layer et update la layer
                layer[i] = bestLayer
                b.position.x = minX + b.radius
                xLast = b.position.x
                xLayer[layer[i]] = b.position.x + b.radius
                hLayer[layer[i]] = max(hLayer[layer[i]], 2*b.radius)
                
                // Compute its children interval bounds
                if t.children.count <= 0 { continue } // no children
                
                t.computeChildMinXMaxX()
                if t.minChildX < xLastParent + Defaults.ProofTreePDF.branchMargin {
                    hSegment -= 1
                    if hSegment < minHSegment { minHSegment = hSegment }
                } else if b.northAnchor.x < xLastChild + Defaults.ProofTreePDF.branchMargin {
                    hSegment += 1
                    if hSegment > maxHSegment { maxHSegment = hSegment }
                }
                t.segmentChildY = CGFloat(hSegment)
                xLastParent = b.northAnchor.x
                xLastChild = t.maxChildX
            }
            
            // Place the segments on the Y axis
            for i in 0..<lev.count {
                let t = lev[i]
                t.segmentChildY = y + (t.segmentChildY - CGFloat(minHSegment)) * Defaults.ProofTreePDF.branchMargin
            }
            y += CGFloat(maxHSegment - minHSegment + 2) * Defaults.ProofTreePDF.branchMargin
            
            // Compute the topY and height of each layer
            for j in stride(from: nbLayers-1, through: 0, by: -1) {
                y += hLayer[j]
                hLayer[j] = y
                y += Defaults.ProofTreePDF.interLayerMargin
            }
            y -= Defaults.ProofTreePDF.interLayerMargin
            
            hLevels[l] = hLayer[layer[0]] - yLevels[l]
            yLevels[l] = hLayer[layer[0]]
            
            // Compute the Y's of the balls
            for i in 0..<lev.count {
                let b = lev[i].pathBall
                b.position.y = hLayer[layer[i]] - b.radius
            }
            
            // Next Level Up
            y += Defaults.ProofTreePDF.interlevelMargin
        }
    }
    
    func draw(in ctx: __CGContext, view: TriangularGridView, oritatamiSystem os: OritatamiSystem) {
//    - (void)drawIn:(__CGContextRef)ctx view:(TerritoireView *)view oritatamiSystem:(OritatamiSystem *)oritatamiSystem startingPoint:(IntPoint)start indexStartingPoint:(int)startIndex slow:(BOOL)slow shiftId:(int)shiftId

        guard let grid = view.grid else { return }
        
        let currentCtx = __GraphicsGetCurrentContext()
        if currentCtx != ctx {
            __GraphicsPushContext(ctx)
        }
        
        // Draw all environments names
        for l in 0..<levels.count {
            let level = levels[l]
            guard level.count > 0 else { continue }
            let i = level[l].pathBall.transcriptIndex
            let bId = os.beadTypeId(indexed: i)
            let bead = os.beadType[bId]
            let envNumber = String(format:"# %04d", environments[l].id)
            let beadName = "Bead " + bead.name
            
            draw(text: envNumber, at: .top,
                 ballPosition: CGPoint(x: -Defaults.ProofTreePDF.environmentMargin/2,
                                       y: yLevels[l] - hLevels[l]),
                 radius: 0,
                 font: Defaults.ProofTreePDF.environmentFont,
                 fillColor: Defaults.ProofTreePDF.environmentColor,
                 strokeColor: __Color.white)

            let attEnv = [NSAttributedString.Key.font : Defaults.ProofTreePDF.environmentFont]
            let dy = NSAttributedString(string: envNumber,
                                        attributes:attEnv).size().height
            
            draw(text: beadName, at: .top,
                 ballPosition: CGPoint(x: -Defaults.ProofTreePDF.environmentMargin,
                                       y: yLevels[l] - hLevels[l] - dy),
                 radius: 0,
                 font: Defaults.ProofTreePDF.environmentFont,
                 fillColor: bead.color,
                 strokeColor: __Color.white)
        }
        if environments.count >= levels.count {
            let nextEnvNumber = String(format: "Next: # %04d", environments.last!.id)
            draw(text: nextEnvNumber,
                 at: .top,
                 ballPosition: CGPoint(x: -Defaults.ProofTreePDF.environmentMargin/2,
                                       y: yLevels[levels.count-1] - hLevels[levels.count-1]),
                 radius: 0,
                 font: Defaults.ProofTreePDF.environmentFont,
                 fillColor: Defaults.ProofTreePDF.environmentColor,
                 strokeColor: __Color.white)
            
        }
    
        // Draw all the segments
        for l in 0..<levels.count-1 {
            for t in levels[l] {
                if t.children.count > 0 {
                    t.computeChildMinXMaxX()
                    let s = t.pathBall.southAnchor
                    let rad = 1.5 * Defaults.ProofTreePDF.handleRadius
                    let r = s + rad * CGRect(x: -1, y: -1, width: 2, height: 2)
                    ctx.setFillColor(Defaults.ProofTreePDF.cropCircleColor.cgColor)
                    ctx.fillEllipse(in: r)
                    
                    ctx.setStrokeColor(Defaults.ProofTreePDF.cropCircleColor.cgColor)
                    let south = t.pathBall.southAnchor
                    let y = t.segmentChildY
                    if t.hasFavorableChild {
                        ctx.setLineCap(.round)
                        ctx.setLineWidth(Defaults.ProofTreePDF.favorableCropCircleLineWidth)
                        let left = min(south.x, t.minFavorableChildX)
                        let right = max(south.x, t.maxFavorableChildX)
                        ctx.strokeLineSegments(between: [
                            south,
                            CGPoint(x: south.x, y: y),
                            CGPoint(x: left, y: y),
                            CGPoint(x: right, y: y)])
                        
                        ctx.setLineWidth(Defaults.ProofTreePDF.cropCircleLineWidth)
                        if right < t.maxChildX {
                            ctx.strokeLineSegments(between: [
                                CGPoint(x: right, y: y),
                                CGPoint(x: t.maxChildX, y: y)
                            ])
                        }
                        if left > t.minChildX {
                            ctx.strokeLineSegments(between: [
                                CGPoint(x:left, y: y),
                                CGPoint(x: t.minChildX, y: y)
                            ])
                        }
                    } else {
                        ctx.setLineWidth(Defaults.ProofTreePDF.cropCircleLineWidth)
                        ctx.strokeLineSegments(between: [
                            south,
                            CGPoint(x: south.x, y: y),
                            CGPoint(x: min(south.x, t.minChildX), y: y),
                            CGPoint(x: max(south.x, t.maxChildX), y: y)
                        ])
                    }
    
                    // Segment to the children
                    for c in t.children {
                        ctx.setLineWidth(c.isFavorable ?
                            Defaults.ProofTreePDF.favorableCropCircleLineWidth :
                            Defaults.ProofTreePDF.cropCircleLineWidth)
                        let north = c.pathBall.northAnchor
                        ctx.strokeLineSegments(between: [
                            north,
                            CGPoint(x: north.x, y: y)
                        ])
                        
                        let rad = 1.5 * Defaults.ProofTreePDF.handleRadius
                        let r = north + rad * CGRect(x: -1, y: -1, width: 2, height: 2);
                        ctx.setFillColor(Defaults.ProofTreePDF.cropCircleColor.cgColor)
                        ctx.fillEllipse(in: r)
                    }
                }
            }
        }
    
        // The balls
        let firstBall = (levels[0][0]).pathBall
        var p = firstBall.origin
        var j = firstBall.transcriptIndex
    
        //NSLog(@"BALL TREE PATH: %@", path);
    
        for l in 0..<levels.count {
            guard levels[l].count > 0 else { continue }
            
            for t in levels[l] {
                // Draw Nb Bonds
                let b = t.pathBall
                let nbBondsText = String(b.nbBonds)
                draw(text: nbBondsText,
                     at: .topLeft,
                     ballPosition: b.position,
                     radius: b.radius,
                     font: t.isFavorable ?
                        Defaults.ProofTreePDF.favorableBondNumberFont :
                        Defaults.ProofTreePDF.bondNumberFont,
                     fillColor: t.isDeadEnd ?
                        Defaults.ProofTreePDF.deadEndColor :
                        __Color.black,
                     strokeColor: __Color.white)
                
                let nbPathsText = String(b.paths.count)
                    draw(text: nbPathsText,
                         at: .bottomRight,
                         ballPosition: b.position,
                         radius: b.radius,
                         font: Defaults.ProofTreePDF.bondNumberFont,
                         fillColor: Defaults.ProofTreePDF.numberOfPathsColor,
                         strokeColor: __Color.white)
            
                // The ball
                ctx.setFillColor(__Color.white.cgColor)
                ctx.fillEllipse(in: b.rect)
                
                b.draw(in: ctx, view: view, oritatamiSystem: os)
                
                view.setNeedsDisplay()
                
                //NSLog(@"Draw:%@ in %@", NSStringFromRect(t.ball.ballRect), ctx);
                
                let firstMove = b.paths[0][0]
                let isSelected = l < transcriptPath.count && firstMove == transcriptPath[l]
                if isSelected && !t.isFavorable {
                    ctx.setStrokeColor(__Color(white: 0.5, alpha: 1).cgColor)
                } else {
                    ctx.setStrokeColor(__Color.black.cgColor)
                }
                
                let lineWidth = t.isFavorable || isSelected ?
                    Defaults.ProofTreePDF.favorableCropCircleLineWidth :
                    Defaults.ProofTreePDF.cropCircleLineWidth
                ctx.setLineWidth(lineWidth)
                if t.isDeadEnd { // dotted line
                    ctx.setLineDash(phase: 0, lengths: [3 * lineWidth,2 * lineWidth])
                    ctx.setLineCap(.butt)
                }
                ctx.strokeEllipse(in: b.rect)
                if t.isDeadEnd { // Undo dotted line
                    ctx.setLineDash(phase: 0, lengths: [])
                    ctx.setLineCap(.round)
                }
            }
            
            j += 1
            if l < transcriptPath.count {
                let d = transcriptPath[l]
                let q = p.neighbor(d)
                grid.synchronize {
//                    let _ = grid.addBead(typeId: os.beadTypeId(indexed:j), at: q, bond: .bond)
                    let _ = grid.addBead(typeId: os.beadTypeId(indexed:j),
                                         from: p, to: d,
                                         isTest: false,
                                         isSeed: j < os.seedPath.count)
                    grid.setBeadIndex(at: q, value: j)
//                    grid.setOutEdge(at: p, to: d)
//                    grid.setBond(at: p, to: d, value: j < os.seedPath.count ? .seed : .primary)
                }
                p = q // move position
            }
        }
        
        if (currentCtx != ctx) {
            __GraphicsPopContext()
        }
    }
    
    
    var englobingRect : CGRect {
        var minX = CGFloat.infinity
        var minY = CGFloat.infinity
        var maxX = -CGFloat.infinity
        var maxY = -CGFloat.infinity
    
        for level in levels {
            for t in level {
                let b = t.pathBall
                
                let minx = b.position.x - b.radius
                let miny = b.position.y - b.radius
                let maxx = b.position.x + b.radius
                let maxy = b.position.y + b.radius
                
                if minx < minX { minX = minx }
                if miny < minY { minY = miny }
                if maxx > maxX { maxX = maxx }
                if maxy > maxY { maxY = maxy }
            }
        }
    
        return CGRect(x: minX, y: minY, width: maxX - minX, height: maxY - minY)
    }
    
    private enum Corner { case topLeft, topRight, bottomLeft, bottomRight, top, bottom }
    
    private func draw(text: String, at corner: Corner, ballPosition center: CGPoint = CGPoint.zero, radius: CGFloat, font: __Font, fillColor: __Color = __Color.black, strokeColor: __Color = __Color.white) {
        let size = text.size()
        let h = size.height
        let w = size.width
        var p = CGPoint.zero
        let r = radius + Defaults.ProofTreePDF.ballTextMargin
        let sa = (r-h)/r
        let ca = cos(asin(sa))
        
        switch corner {
        case .topLeft:
            p = CGPoint(x: -w - r * ca, y: r - h)
        case .topRight:
            p = CGPoint(x: r * ca, y: r - h)
        case .bottomLeft:
            p = CGPoint(x: -w - r * ca, y: -r + h)
        case .bottomRight:
            p = CGPoint(x: r * ca, y: -r + h)
        case .top:
            p = CGPoint(x: r - w/2, y: r - h)
        case .bottom:
            p = CGPoint(x: r - w/2, y: -r + h)
        }

        let fillAttribute : [NSAttributedString.Key: Any] = [
            .font: font,
            .foregroundColor: fillColor
        ]
        let outlineAttribute : [NSAttributedString.Key: Any] = [
            .font: font,
            .foregroundColor: __Color.clear,
            .strokeColor: __Color.white,
            .strokeWidth: font.pointSize*2,
        ]

        let outlineText = NSAttributedString(string: text, attributes: outlineAttribute)
        outlineText.draw(at: p)
        
        let fillText = NSAttributedString(string: text, attributes: fillAttribute)
        fillText.draw(at: p)
    }
    
}
