//
//  Defaults.swift
//  Oritatami Builder
//
//  Created by Nikao on 10/01/2019.
//  Copyright © 2019 Nikao. All rights reserved.
//

import Foundation
import QuartzCore
import UIKit

let os_file_extension = "os"
let os_file_extensions = ["os", "OS", "o_s", "O_S"]

extension BeadTypeSet {
    struct Defaults {
        static let maxSize : Int = 4000
    }
}

extension BeadType {
    static let empty_beadType_id = -1
    struct Defaults {
        static let nameFixedLength : Int = 4
    }
}

extension CGRect {
    struct Defaults {
        static let intPointRadius: CGFloat = 0.5
    }
}

extension TriangularGrid {
    struct Defaults {
        static let width = 20
        static let height = 20
    }
}

extension TriangularGridView {
    struct Defaults {
        static let minScale: CGFloat = 0.05
        static let minDrawBeadScale: CGFloat = 8
        static let minDrawGridScale: CGFloat = 20
        static let bondLength: CGFloat = 40
        static let scale: CGFloat = bondLength/1.4 // minDrawBeadScale
        static let primaryBondColorUpfront: __Color = __Color(white:0, alpha:0.4)
        static let primaryBondColorUpfrontPDF: __Color = __Color(white:0, alpha:1)
        static let primaryBondColorBehind: __Color = __Color(white:0, alpha:1)
        static let primarySeedBondColor: __Color = __Color.brown500 // __Color(white:0.25, alpha:1)
        static let primarySeedBondLineWidth: CGFloat = 1.5*primaryBondLineWidthBehind
        static let primaryTestBondColor: __Color = __Color.lightBlue400 // __Color(white:0.25, alpha:1)
        enum BondStyle { case lineWidth, strokeColor, dashPattern }
        static let bondStyle = { (b: Bond) -> (lineWidth: CGFloat, strokeColor: __Color, dashPattern: [CGFloat]) in
            switch b {
            case .bond:
                return (secondaryBondLineWidth, secondaryBondColor, secondaryBondLineDash)
            case .testBond:
                return (secondaryTestBondLineWidth, secondaryBondColor, secondaryBondLineDash)
            case .testPrimary, .testSeed:
                return (primaryTestBondLineWidthBehind, .black, [])
            case .primaryBestSoFar:
                return (primaryTestBondLineWidthBehind, primaryBestSoFarBondColor, [])
            case .primaryBest:
                return (primaryTestBondLineWidthBehind, primaryBestBondColor, [])
            case .seed:
                return (primarySeedBondLineWidth, primarySeedBondColor, [])
            case .primary:
                return (primaryBondLineWidthBehind, .black, [])
                
            }
        }
        static let min_compressed_path_line_width : CGFloat = 1
        static let primaryBestSoFarBondColor: __Color = __Color.purpleA200 // lightGreenA400
        static let primaryBestBondColor: __Color = __Color.amber600
        static let primaryBondLineWidthUpfront: CGFloat = 8/Defaults.bondLength
        static let primaryBondLineWidthBehind: CGFloat = 8/Defaults.bondLength
        static let primaryTestBondLineWidthUpfront: CGFloat = 2*primaryBondLineWidthBehind
        static let primaryTestBondLineWidthBehind: CGFloat = primaryTestBondLineWidthUpfront
        static let secondaryBondColor: __Color = __Color.black
        static let secondaryBondLineWidth: CGFloat = 2/Defaults.bondLength
        static let secondaryTestBondLineWidth: CGFloat = primaryTestBondLineWidthBehind
        static let secondaryBondLineDash: [CGFloat] = [5/Defaults.bondLength,3/Defaults.bondLength]
        static let upscaledGridColor: __Color = __Color.orange
        static let gridColor: __Color = __Color.cyan200
        static let gridLineWidth: CGFloat = 0.75/Defaults.bondLength
        static let upscaledGridLineWidth: CGFloat = 1/Defaults.bondLength
        static let beadRadius: CGFloat = 1/4
        static let unoccupiedBeadRadius: CGFloat = beadRadius/5
        static let beadInnerRadius: CGFloat = Defaults.beadRadius*3/5
        static let beadColors: [__Color] = [__Color.red, __Color.orange500, __Color.yellow500, __Color.green500, __Color.lightBlue500, __Color.indigo500, __Color.purple500, __Color.black, __Color.cyan]
        static let beadLineColor: __Color = __Color.black
        static let beadLineWidth: CGFloat = 1/Defaults.bondLength
        static let blockedBeadColor: __Color = __Color.red
        static let anomalyBeadColor: __Color = __Color.yellow
        static let arrowProbability: Double = 0.3
        static let hexagonMinBrightness: CGFloat = 0.9
        static let hexagonMaxBrightness: CGFloat = 0.95
        static let hueStep: CGFloat = 1.0/256
        
        // Pattern drawing
        static let patternPath: String = "patterns/"
        static let patternPDFscale: CGFloat = 10
        
        static let patternGridExtraLength: CGFloat = 0.8
        
        static let patternWallColor: __Color = __Color(red: 0, green: 0.9, blue: 0, alpha: 1)
        static let patternWallFillColor: __Color = __Color(red: 0, green: 1, blue: 0, alpha: 0.25)
        static let patternPluggingWallColor: __Color = __Color(red: 0.5, green: 0, blue: 1, alpha: 1)
        static let patternPluggingWallFillColor: __Color = __Color(red: 0.5, green: 0, blue: 1, alpha: 0.25)
        static let patternPluggingArrowColor: __Color = __Color(red:229/255, green: 78/255, blue: 250/255, alpha: 1) // __Color(white:0.5, alpha:1) //__Color(red: 0.75, green: 0.15, blue: 1, alpha: 1)
        static let patternWallThickness: CGFloat = 8/Defaults.bondLength
        static let patternWallExtend: CGFloat = 0.5
        
        static let patternBondColor: __Color = __Color(white: 0.25, alpha: 1)
        static let patternBondThickness: CGFloat = 2/Defaults.bondLength
        static let patternPathColor: __Color = __Color.black
        static let patternPathThickness: CGFloat = 8/Defaults.bondLength
        static let patternPathArrowColor: __Color = __Color.red
        static let patternPathSeedColor: __Color = __Color(red: 1, green: 0.75, blue: 0, alpha: 1)
        static let patternPathSeedThickness: CGFloat = 16/Defaults.bondLength
        static let patternTimeAnomalyColor: __Color = __Color(red: 1, green: 0.6, blue: 0.6, alpha: 1) // //__Color(red: 1, green:1, blue: 0, alpha: 1)
        static let patternPathAnomalyColor: __Color = __Color(red: 1, green: 0.8, blue: 0.05, alpha: 1) // __Color(red: 150/255, green: 240/255, blue: 1, alpha: 1) //__Color(red: 1, green:1, blue: 0, alpha: 1)
        static let patternAnomalyRadius: CGFloat = 1/2
        static let patternFontName: String = "CMU Serif"
        static let patternFontSize: CGFloat = 0.25
        
        static let transcriptMaxNumberOfBeadsPerRow = 10
        
        // Basic Bead Type
        static let basicBeadTypeRadius : CGFloat = 1/4
        static let basicBeadTypeArmRadius : CGFloat = basicBeadTypeRadius/3
        static let basicBeadTypeArmExtendedLength : CGFloat = 1/2
        static let basicBeadTypeArmRetractededLength : CGFloat = (basicBeadTypeArmExtendedLength + basicBeadTypeRadius)/2
        
        // MapOrigin moving
        static let mapOriginMoveLength : TimeInterval = 0.75
        
        // debug
        static let show_debug_grid_dictionnary = false // true
    }
}

extension TriangularGridViewController {
    struct Defaults {
        static let demoSleepTime = 2.0
        static let updateRadius: CGFloat = 5 // 5 Needed for fixing anomalies
        static let frameRate: Double = frameRateFast // 1/500 // 1/1000
        static let frameRateSlow: Double = 1/200 // 1/1000
        static let frameRateFast: Double = 1/3000 // 1/1000
        static let infoLabelShowTimeInterval: TimeInterval = 2
        static let infoLabelHideTransition: TimeInterval = 0.5
        static let infoLabelTextColor = __Color.white
        static let infoLabelBackgroundColor = __Color(white: 0, alpha: 0.5)
        
        static let gifFoldingStepRootFilename: String = "folding"
        static let gifRunFoldingRootFilename: String = "run"
    }
    
    struct Alerts {
        static let stopSimulationFirst = "Stop simulation first!"
        static let stopingSimulationFirst = "Stoping simulation first!"
    }
}

extension OpportunisticOritatamiSimulator {
    struct Defaults {
        static let updateRadius: CGFloat = 2
        static let frameRate: Double = 0 // 1/500// 1/100
        
        static let gifTestFrameRate: CGFloat = 2*CGFloat(frameRate)
        static let gifShuffleTestFrameRate: CGFloat = CGFloat(frameRate)/2
        static let gifLoopBestMaxFrameRate: CGFloat = 2
        static let gifLoopBestTotalFrameRate: CGFloat = 5
        static let gifLoopBestMaxNumberOfConfigurations: Int = 20
        
    }
}

extension OpportunisticOritatamiSimulatorRemember {
    struct Defaults {
        static let updateRadius: CGFloat = 2
        static let frameRate: Double = 0 // 1/500// 1/100
        
        static let gifTestFrameRate: CGFloat = 2*CGFloat(frameRate)
        static let gifShuffleTestFrameRate: CGFloat = CGFloat(frameRate)/2
        static let gifLoopBestMaxFrameRate: CGFloat = 2
        static let gifLoopBestTotalFrameRate: CGFloat = 5
        static let gifLoopBestMaxNumberOfConfigurations: Int = 20
        
        
        static let use_local_env = true
        static let print_local_env = false
        static let print_compressed_path = false
    }
}

extension OritatamiSystem {
    struct Defaults {
        static let name = "Oritatami system"
        static let periodCount = 1 // one period by default
    }
}

extension TriangularGridDictionary {
    struct Defaults {
        // MARK: dictionary
        static let dict_bitsize = 6 // 9

        // MARK: cache: 4x4 cells
        static let cache_cell_bitsize = 3 // allow upto delay≤6
        
        // MARK: uses cache by default
        static let read_write_from_cache_by_default = false
        
        // MARK: auto-push
        static let auto_push = true
        static let push_includes_neighbors = true
        
        // MARK: auto-purge
        static let auto_flush_surrounded_beads = false
        static let nb_steps_between_flushes_of_surrounded_beads : Int = 100_000
    }
}

