//
//  LocalEnvironment.swift
//  Simple OS Simulator
//
//  Created by Nikao on 29/05/2021.
//  Copyright © 2021 Nikao. All rights reserved.
//

import Foundation

// Il faut ajouter les δ nascent beads à l'environnement !!!!
struct LocalEnvironment : Hashable {
    let radius : Int
    let hex_count : Int
    var beads : [Int]
    
    init(radius r: Int) {
        radius = r
        hex_count = 1 + 3*r*(r+1)
        beads = Array(repeating: BeadType.empty_beadType_id,
                      count: hex_count + r-1) // r-1 = les r-1 nascent beads
    }
    
    // Index of points inside the environment

    func hex_index(_ p: IntPoint) -> Int {
        let idx = (p.x * radius + p.y * (radius+1)) % hex_count
        return idx >= 0 ? idx : (idx + hex_count)
    }

    // init from a grid + beadtypestream

    init(at p: IntPoint, in grid: TriangularGridDictionary, delay: Int, nascentBeads: [BeadTypeId],
         readFromCacheWithoutChecking rfcwc: Bool = true) {
        radius = delay + 1
        hex_count = 1 + 3*radius*(radius+1)
        beads = Array(repeating: BeadType.empty_beadType_id, count: hex_count + delay)
        
        var k = 0
        for j in -radius...radius {
            for i in max(-radius+j,-radius)...min(radius+j,radius) {
                let q = p + IntPoint(x:i, y: j)
//                beads[index(q)] = grid.beadTypeId(at: q, readFromCacheWithoutChecking: rfcwc) ?? BeadType.empty_beadType_id
                beads[k] = grid.beadTypeId(at: q, readFromCacheWithoutChecking: rfcwc) ?? BeadType.empty_beadType_id
                k += 1
            }
        }
        for i in 0..<min(delay, nascentBeads.count) {
            beads[k] = nascentBeads[i]
            k += 1
        }
    }

    // Move in a grid -> Il faut réécrire cette procédure en utilisant une permutation qui va chercher dans le cache seulement quand c'est nécessaire
    mutating func move(at p: IntPoint, to d: Direction, in grid: TriangularGridDictionary,
                       readFromCacheWithoutChecking rfcwc: Bool = true) {
        var q = p + d.intPoint + radius * d.ccw.intPoint
        beads[hex_index(q)] = grid.beadTypeId(at: q, readFromCacheWithoutChecking: rfcwc) ?? BeadType.empty_beadType_id
        for dd in [d.cw, d.cw.cw] {
            for _ in 0..<radius {
                q = q.neighbor(dd)
                beads[hex_index(q)] = grid.beadTypeId(at: q, readFromCacheWithoutChecking: rfcwc) ?? BeadType.empty_beadType_id
            }
        }
    }

    // Hashable
    
    func hash(into hasher: inout Hasher) {
        for b in beads {
            hasher.combine(b)
        }
    }
    
    // Equatable
    
    static func == (lhs: LocalEnvironment, rhs: LocalEnvironment) -> Bool {
        return lhs.beads == rhs.beads
    }
    
    // pretty print
    func prettyPrint(with bts: BeadTypeSet) -> String {
        var s = ""
        var k = 0
        for j in -radius...radius {
            s += String(repeating: "  ", count: abs(j))
            for _ in max(-radius+j,-radius)...min(radius+j,radius) {
                let b = beads[k]
                let sb = b >= 0 ? bts.beadType[b].fixedLengthName : " °  "
                k += 1
                s += "\(sb)"
            }
            s += "\n"
        }
        s += "Nascents: "
        for _ in 0..<radius-1 {
            let b = beads[k]
            let sb = b >= 0 ? bts.beadType[b].fixedLengthName : " °  "
            k += 1
            s += "\(sb)"
        }
        return s + "\n"
    }
}
