//
//  OpportunisticOritatamiSimulator
//  Scary Pacman 2
//
//  Created by Nikao on 16/3/18.
//  Copyright © 2018 Algoseum. All rights reserved.
//

import UIKit

class OpportunisticOritatamiSimulator : OritatamiSimulator {
    
    var grid: TriangularGridDictionary //TriangularGridProtocol
    
    var oritatamiSystem: OritatamiSystem {
        didSet {
            grid.beadTypeSet = oritatamiSystem.beadTypeSet
            grid.rule = oritatamiSystem.rule
            
            
            grid.auto_push = true
            grid.push_includes_neighbors = true
        }
    }
    var beadTypeSet: BeadTypeSet { return oritatamiSystem.beadTypeSet }
    private var beadTypeIdStream: [BeadTypeId] { return oritatamiSystem.beadTypeIdsSequence } // Deprecated
    var seedPath: [Direction] { return oritatamiSystem.seedPath }
    var delay: Int { return oritatamiSystem.delay }
    var beadStreamCount: Int { return oritatamiSystem.beadSequenceCount }
    
    var currentPosition = IntPoint.zero
    var currentIndex: Int = 0
    
    weak var view: TriangularGridView? = nil
    var frameRate: Double = Defaults.frameRate
    var animated : Bool = true
    var follow : Bool = true
    
    var saveAsGif: Bool = false {
        didSet {
            guard saveAsGif != oldValue else { return }
            if !saveAsGif && runTimedImageSequence.count > 0 {
                let _ = saveGif(&runTimedImageSequence, rootfilename: gifRunRootFilename)
            }
        }
    }
    var shuffleGif: Bool = false
    var noLoopOnBest: Bool = true
    var loopOnNonDeterminism: Bool = false
    var gifFoldingStepRootFilename: String = "folding"
    var gifRunRootFilename: String = "OSsimulatorRun"
    var foldingTimedImageSequence : TimedImageSequence = []
    var runTimedImageSequence : TimedImageSequence = []

    var currentBestFirstStep: [Direction] = []
    var currentBestNumberOfBonds: Int = 0
    
    var mustStop = false
    
    required init(width w: Int, height h: Int, oritatamiSystem os: OritatamiSystem) {
        grid = TriangularGridDictionary() //TriangularGrid(width: w, height: h)
        oritatamiSystem = os
        grid.beadTypeSet = os.beadTypeSet
        grid.rule = os.rule
    }

//    func centerIfNeeded() {
//        guard let v = view else { return }
//        let b = v.bounds
//        let viewableArea = b.insetBy(dx: 0.2*b.width, dy: 0.2*b.height)
//        let p = v.point(for: currentPosition)
//        if !viewableArea.contains(p) {
//            v.center(at: currentPosition.point)
//        }
//    }

    var bestNumberOfBonds = 0
    var bestDir : [ Direction ] = []
    var isCurrentBestDir : Bool = false
    var bestPathsSoFar : [Path] = []
    var currentPath : Path = []
    var currentNumberOfBonds = 0

    @inline(__always) func beadTypeId(indexed j: Int) -> BeadTypeId {
        return oritatamiSystem.beadTypeId(indexed: j)
    }

    
    func nextStep(index i: Int, lastPosition p: IntPoint) -> Int {
        guard i <= delay else { return 0 }
        
        let j = i + currentIndex
        guard j < beadStreamCount else { return -1 }
        
        let isLastStep = (i == delay || j == beadStreamCount-1)
        
        // Look for initial direction
        let d_init : Direction = .NE
        var d0 = d_init
        while !grid.isOccupied(at: p.neighbor(d0)) && d0 != d_init.ccw {
            d0 = d0.cw
        }
        
//        print("d0: \(d0) - i: \(i)")
        var d = d0.cw
        for _ in Direction.all {
            let q = p.neighbor(d)
            guard !grid.isOccupied(at: q) else { d = d.cw; continue }
            
            if i == 1 {
                currentPath = [d]
                isCurrentBestDir = false
            } else {
                currentPath.append(d)
//                print("\(i): \(currentPath)")
            }
            
            var nbNewBonds : Int = 0
            
            grid.synchronize {
                nbNewBonds = grid.addBead(typeId: beadTypeId(indexed: j), at: q, bond: .testBond, writeToCacheWithoutChecking: true)
                grid.setOutEdge(at: p, to: d, writeToCacheWithoutChecking: true)
                grid.setBond(at: p, to: d, value: .testPrimary, writeToCacheWithoutChecking: true)
            }

            currentNumberOfBonds += nbNewBonds
            
            if !isLastStep {
                let _ = nextStep(index: i+1, lastPosition: q)
            } else {
                var pathColor = TriangularGridView.Defaults.primaryTestBondColor
                var delay = (shuffleGif ?  Defaults.gifShuffleTestFrameRate : Defaults.gifTestFrameRate)
                
//                print("Current path \(currentPath) from \(currentPosition)")
                
                if currentNumberOfBonds > bestNumberOfBonds {
                    bestNumberOfBonds =  currentNumberOfBonds
                    bestDir = []
                    bestPathsSoFar = []
//                    let v = view
//                    let info = "\(currentNumberOfBonds) bond\(currentNumberOfBonds >= 2 ? "s": "")"
//                    DispatchQueue.main.async {
//                        v?.controller?.show(info: info)
//                    }
                    if !shuffleGif {
                        pathColor = TriangularGridView.Defaults.primaryBestSoFarBondColor // __Color(white: 0.35, alpha: 1) // __Color(hue: CGFloat(currentNumberOfBonds)/5, saturation: 1, brightness: 1, alpha: 1)
                        delay *= 3
                    }
              }
                
                if currentNumberOfBonds >= bestNumberOfBonds {
                    isCurrentBestDir = true
//                    print("Added: \(currentPath)")
                    bestPathsSoFar.append(currentPath)
                    
//                    pathColor = __Color.lightBlue500 // __Color(white: 0.35, alpha: 1) // __Color(hue: CGFloat(currentNumberOfBonds)/5, saturation: 1, brightness: 1, alpha: 1)
//                    delay *= 2
                }

                let v = view
                let dp = currentPosition.point
                let r = CGFloat(delay)+2
                
                v?.primaryTestBondColor = pathColor
                DispatchQueue.main.async {
                    v?.setNeedsDisplay(inUnscaled:CGRect(circleAt: dp, radius:r))
                }
                
                if animated && saveAsGif {
                    recordView(delay: delay, pathColor: pathColor, in: &foldingTimedImageSequence)
                }
                
                if animated && !saveAsGif {
                    Thread.sleep(forTimeInterval: frameRate)
                }
                
            }
            
            grid.synchronize {
                grid.removeBead(at: q)
                DispatchQueue.main.async {
                    self.view?.setNeedsDisplay(inUnscaled:CGRect(circleAt: q.point, radius:1))
                }
            }
            currentPath.removeLast()
            currentNumberOfBonds -= nbNewBonds

            if i == 1 && isCurrentBestDir {
                bestDir.append(d)
            }
            
            d = d.cw
        }
        
        if i == 1 {
            // Best Paths
//           print("Best Dirs: \(bestDir)\nBest paths: \(bestPathsSoFar)\n")
            
            if shuffleGif && !foldingTimedImageSequence.isEmpty {
                foldingTimedImageSequence.shuffle()
            }
            
            if !noLoopOnBest {
                // Loop on best paths
                loopOnBestPath()
            }
            
            // Check for determinism
            if bestDir.count == 0 {
                print(" No extension!")
                return -1
            } else if bestDir.count == 1 {
                let d = bestDir[0]
//                print("Next direction is \(d)")
                // Make the step
                grid.synchronize {
                    currentPosition = p.neighbor(d)
                    currentIndex += 1
                    let _ = grid.addBead(typeId: beadTypeId(indexed:currentIndex), at: currentPosition, bond: .bond, writeToCacheWithoutChecking: true)
                    grid.setBeadIndex(at: currentPosition, value: currentIndex, writeToCacheWithoutChecking: true)
                    grid.setOutEdge(at: p, to: d, writeToCacheWithoutChecking: true)
                    grid.setBond(at: p, to: d, value: .primary, writeToCacheWithoutChecking: true)
                
                    grid.commit(point: p)
                    grid.shift_cache(to_include_in_center: p)
                }
                
                let q = currentPosition.point
                DispatchQueue.main.async { [weak self] in
                    guard let me = self else { return }
                    if me.follow { me.view?.centerIfNeeded(at: me.currentPosition) }
                    me.view?.setNeedsDisplay(inUnscaled:CGRect(circleAt: q, radius:1))
                }

                
                return 0
            } else {
                print("Non-determinism! \(bestDir)")
                DispatchQueue.main.async { [weak self] in
                    guard let me = self else { return }
                    if me.follow { me.view?.centerIfNeeded(at: me.currentPosition) }
                    if let ctrl = me.view?.controller {
                        ctrl.show(info:"Non-determinism! \(me.bestDir)")
                        if !ctrl.show_view {
                            ctrl.show_hide_view(me)
                        }
                    }
                }
                loopOnNonDeterminism = true
                while !mustStop {
                    loopOnBestPath()
                }
                loopOnNonDeterminism = false
                return -2
            }
        }
        
        return bestNumberOfBonds
    }
    
    func flushPendingGif() {
        if runTimedImageSequence.count > 0 {
            saveGif(&runTimedImageSequence, rootfilename: gifRunRootFilename)
        }
    }
    
    func nextStep() -> Bool {
        bestDir = []
        bestNumberOfBonds = 0
        bestPathsSoFar = []
        currentPath = []

        let output = nextStep(index: 1, lastPosition: currentPosition)

        if foldingTimedImageSequence.count > 0 {
            let _ = saveGif(&foldingTimedImageSequence, rootfilename: gifFoldingStepRootFilename)
        }

        if saveAsGif {
            recordView(in: &runTimedImageSequence)
        }
        
//        let envMask = EnvironmentMask(at: currentPosition, lastBeadIndex: currentIndex, grid: grid, oritatamiSystem: oritatamiSystem)
        let currentBeadTypeName = oritatamiSystem.beadType[oritatamiSystem.beadTypeId(indexed: currentIndex)].name
//        print(currentPosition, currentIndex, currentBeadTypeName, "\n", envMask)

        let success = output >= 0

        if !success && runTimedImageSequence.count > 0 {
            let _ = saveGif(&runTimedImageSequence, rootfilename: gifRunRootFilename)
        }
        
        return success
    }

    func loopOnBestPath() {
        guard let v = view else { return }
        v.primaryTestBondColor = TriangularGridView.Defaults.primaryBestBondColor
        let delay = min(Defaults.gifLoopBestMaxFrameRate, Defaults.gifLoopBestTotalFrameRate/CGFloat(bestPathsSoFar.count))
        for path in bestPathsSoFar {
//            print(path)
            if mustStop { return }
            
            grid.synchronize {
                draw(bestpath: path, erase: false)
            }
            DispatchQueue.main.sync { [weak self] in self?.view?.setNeedsDisplay() }
            if saveAsGif {
                if animated {
                    recordView(delay: delay, pathColor: TriangularGridView.Defaults.primaryBestBondColor, in: &foldingTimedImageSequence)
                }
                recordView(pathColor: TriangularGridView.Defaults.primaryBestBondColor,
                           in: &runTimedImageSequence)
            }
            if (animated || loopOnNonDeterminism) && !saveAsGif {
                Thread.sleep(forTimeInterval: TimeInterval(delay))
            }
            grid.synchronize {
                draw(bestpath: path, erase: true)
            }
        }
        v.primaryTestBondColor = TriangularGridView.Defaults.primaryTestBondColor
    }

    func draw(bestpath: Path, erase: Bool = false) {
        var p = currentPosition
        var j = currentIndex + 1
        if !erase {
            for d in bestpath {
                let q = p.neighbor(d)
                let _ = grid.addBead(typeId: beadTypeId(indexed: j), from: p, to: d, isTest: true, isSeed: false, writeToCacheWithoutChecking: true)
                grid.setBond(at: p, to: d, value: .primaryBest)
                //                let _ = grid.addBead(typeId: beadTypeId(indexed: j), at: q, bond: .testBond)
//                grid.setOutEdge(at: p, to: d)
//                grid.setBond(at: p, to: d, value: .primaryBest)
                p = q
                j += 1
            }
        } else {
            for d in bestpath {
                p = p.neighbor(d)
                grid.removeBead(at: p, writeToCacheWithoutChecking: true)
            }
        }
    }
    

    
    func drawSeed(from p: IntPoint) {
//        print("Draw seed: \(seedPath) from \(p)")
        guard beadTypeIdStream.count > seedPath.count else { return }
                
        currentIndex = 0
        currentPosition = p
//        grid.synchronize {
            grid.clear()
            grid.pathStartingPoints.append(currentPosition)
        grid.init_cache(at: currentPosition)
        let _ = grid.addBead(typeId: beadTypeId(indexed: currentIndex), at: currentPosition, bond: .seed, writeToCacheWithoutChecking: true)
            grid.setBeadIndex(at: currentPosition, value: currentIndex, writeToCacheWithoutChecking: true)
            grid.commit(point: p)
            for d in seedPath {
                currentIndex += 1
                let p = currentPosition.neighbor(d)
                let _ = grid.addBead(typeId: beadTypeId(indexed: currentIndex), at: p, bond: .bond, writeToCacheWithoutChecking: true)
                grid.setBeadIndex(at: p, value: currentIndex, writeToCacheWithoutChecking: true)
                grid.setOutEdge(at: currentPosition, to: d, writeToCacheWithoutChecking: true)
                grid.setBond(at: currentPosition, to: d, value: .seed, writeToCacheWithoutChecking: true)
                currentPosition = p
                grid.commit(point: p)
                grid.shift_cache(to_include_in_center: currentPosition)
            }
//        }
        print("Draw seed: Done, ends at \(currentPosition) current index \(currentIndex)")
    }
    
    
    // .MARK: GIF saving
    
    func saveGif(_ imageSequence: inout TimedImageSequence, to url: URL) -> Bool {
        let once : Bool = { print("URL: \(url)"); return true }()
        _ = once
        
        let result = imageSequence.saveAsGif(to: url, loopCount: 0)
        if result {
            imageSequence = []
        }
        return result
    }
    
    private var printFileLocationOnce: Bool = true
    func saveGif(_ imageSequence: inout TimedImageSequence, rootfilename: String) -> Bool {
        let filename = String(format: "%@-%04d.gif", rootfilename, currentIndex)
        guard let url = URL.documentURL(for: filename) else { return false }
        if printFileLocationOnce {
            print("Saving to \(url)...")
            printFileLocationOnce = false
        }
        let result = imageSequence.saveAsGif(to: url, loopCount: 0)
        if result {
            imageSequence = []
        }
        return result
    }
    
    func recordView(delay: CGFloat = Defaults.gifTestFrameRate,
                    pathColor: __Color = TriangularGridView.Defaults.primaryTestBondColor,
                    in sequence: inout TimedImageSequence,
                    hideInfoLabel : Bool = true) {
        guard let v = view else { return }
        var img : __Image? = nil
        DispatchQueue.main.sync {
            let h = v.controller?.infoLabel.isHidden ?? true
            CATransaction.setDisableActions(true)
            v.primaryTestBondColor = pathColor
            v.setNeedsDisplay()
            if hideInfoLabel { v.controller?.infoLabel.isHidden = true }
            img = __Image(view: v)
            if hideInfoLabel { v.controller?.infoLabel.isHidden = h }
            CATransaction.setDisableActions(false)
        }
        if let img = img {
            sequence.append((img, delay))
        }
    }
}

//class PathTree<T> : CustomStringConvertible where T : CustomStringConvertible {
//    var value: T
//    var subtrees: [PathTree<T>] = []
//    weak var parent: PathTree<T>? = nil
//
//    init(value v: T) {
//        value = v
//    }
//
//    func addSubTree(_ t: PathTree<T>) {
//        if let p = t.parent {
//            p.removeSubTree(t)
//        }
//        subtrees.append(t)
//        t.parent = self
//    }
//
//    func removeSubTree(_ t: PathTree<T>) {
//        guard let p = t.parent, self === p else { return }
//        for i in 0..<subtrees.count {
//            if t === subtrees[i] {
//                subtrees.remove(at: i)
//                t.parent = nil
//                return
//            }
//        }
//    }
//
//    func _description(prefix: String = "") -> String {
//        var s = prefix + value.description + "\n"
//        let new = prefix + "  "
//        for t in subtrees {
//            s += t._description(prefix: new)
//        }
//        return s
//    }
//
//    var description: String {
//        return _description()
//    }
//}

