//
//  TriangularGrid.swift
//  Scary Pacman
//
//  Created by Nikao on 19/9/17.
//  Copyright © 2017 Algoseum. All rights reserved.
//


// Séparer en deux class: Triangular grid (contient échelle + routine de base: récupérer du point de cliquage etc.)+ pathconstructor avec une classe pour chaque ensemble d'échelle + peut-être une classe pour mémoriser la forme construire pour pouvoir la passer à différentes échelles.

// Réparer cellNeighbor -> trouver un fix pour les échelles x.5 et l'échelle 3.0, ne pas chercher des fonctions trop générale si pas utile pour échelle autres que 3.0 (cf par exemple la nécessité de regarder les arêtes -> pourquoi ne pas stocker les anomalies au centre de la cellule? est-ce vraiment utile?)

import Foundation
import QuartzCore

class TriangularGrid : TriangularGridProtocol {
        
    static let queueName = "gridQueue"
    let queue = DispatchQueue(label: TriangularGrid.queueName, qos: .userInitiated, attributes: DispatchQueue.Attributes()) // sequential queue
    
    func currentQueueName() -> String? {
        let name = __dispatch_queue_get_label(nil)
        return String(cString: name, encoding: .utf8)
    }
    
//    @objc
    func synchronize(closure: () -> ()) {
        //objc_sync_enter(self)
        queue.sync { closure() }
        //objc_sync_exit(self)
    }
    
    func synchronizeIfNotAlreadySynchronized(closure: () -> ()) {
        if currentQueueName() == TriangularGrid.queueName {
            closure()
        } else {
            queue.sync { closure() }
        }
    }
    
    var rect: IntRect { didSet { updateMapSize() } }
    
    var width: Int { get { return rect.width } }
    var height: Int { get { return rect.height } }
    var origin: IntPoint { get { return rect.origin } }
    
    var minX: Int { return origin.x }
    var minY: Int { return origin.y }
    var maxX: Int { return width - 1  + origin.x }
    var maxY: Int { return height - 1 + origin.y }

    var pathStartingPoints: [ IntPoint ] = []
    
    private var beadTypeId: Array<BeadTypeId?> = []
    private var beadIndex: Array<Int?> = []
    private var bond: Array<Bond?> = []
    private var outEdge: Array<Direction?> = []
    var info: Array<[GridInfo:Any]?> = []

    var beadTypeSet : BeadTypeSet? = nil
    var rule: OritatamiRule? = nil
    
    
    required init(width: Int = Defaults.width, height: Int = Defaults.height) {
        self.rect = IntRect(origin: IntPoint(x: -width/2, y: -height/2), size: IntSize(width: width, height: height))
        updateMapSize()
        clear()
    }

    @inline(__always) func updateMapSize() {
        mapSize = rect.size.width * rect.size.height
    }
    
    func allocArrays() {
        print("Map size: \(mapSize)")
        self.beadIndex = Array(repeating: nil, count: self.mapSize)
        self.outEdge = Array(repeating: nil, count: self.mapSize)
        self.beadTypeId = Array(repeating: nil, count: self.mapSize)
        self.bond = Array(repeating: .none, count: self.mapSize * Direction.halfCount)
        self.info = Array(repeating: nil, count: self.mapSize)
    }
    
    func clear() {
        allocArrays()
        pathStartingPoints = []
    }
    
    var mapSize: Int = 0

    var boundingRect: CGRect {
        return CGRect(x: origin.point.x+3, y: origin.point.y+3,
                      width: CGFloat(width-6), height: IntPoint.sqrt32 * CGFloat(height-6))
    }
    
    @inline(__always) func contains(_ p: IntPoint) -> Bool {
        let x = p.x - origin.x
        let y = p.y - origin.y
        
        return x >= 0 && x < width && y >= 0 && y < height
    }

    @inline(__always) func index(_ p: IntPoint) -> Int? {
        let x = p.x - origin.x
        let y = p.y - origin.y
        
        if x >= 0 && x < width && y >= 0 && y < height {
           return y * width + x
        } else {
            return nil
        }
    }
    
    @inline(__always) func unsafeIndex(_ p: IntPoint) -> Int {
        let x = p.x - origin.x
        let y = p.y - origin.y
        
        return y * width + x
    }

    func isOccupied(at p:IntPoint) -> Bool {
        if let i = index(p) {
            return beadTypeId[i] != nil
        }
        return false
    }
    
    func beadTypeId(at p:IntPoint) -> BeadTypeId? {
        if let i = index(p) {
            return beadTypeId[i]
        }
        return nil
    }
    
    func setBeadTypeId(at p:IntPoint, value: BeadTypeId?) {
        if let i = index(p) {
            beadTypeId[i] = value
        } else {
            print("Trying to set occupied at \(p) outside the map!!!")
        }
    }
    
    func beadIndex(at p:IntPoint) -> Int? {
        if let i = index(p) {
            return beadIndex[i]
        }
        return nil
    }
    
    func setBeadIndex(at p:IntPoint, value: Int?) {
        if let i = index(p) {
            beadIndex[i] = value
        } else {
            print("Trying to set timeStamp at \(p) outside the map!!!")
        }
    }
    

    @inline(__always) func bondIndex(_ p: IntPoint, _ d: Direction) -> Int? {
        if d.rawValue < Direction.halfCount {
            if let i = index(p) {
                return i * Direction.halfCount + d.rawValue
            }
        } else {
            if let i = index(p.neighbor(d)) {
                return i * Direction.halfCount + (-d).rawValue
            }
        }
        return nil
    }
    
    func bond(at p: IntPoint, direction d: Direction) -> Bond? {
        if let i = bondIndex(p, d) {
            return bond[i]
        }
        return nil
    }
    
    func setBond(at p: IntPoint, to d: Direction, value: Bond?) {
        if let i = bondIndex(p, d) {
            bond[i] = value
        } else {
            print("Trying to set bond at \(p) outside the map!!!")
        }
    }

    func outEdge(at p:IntPoint) -> Direction? {
        if let i = index(p) {
            return outEdge[i]
        }
        return nil
    }

    func setOutEdge(at p: IntPoint, to d: Direction?) {
        if let i = index(p) {
            outEdge[i] = d
        } else {
            print("Trying to set outEdge at \(p) outside the map!!!")
        }
    }
    
    @inline(__always) func addBead(typeId: BeadTypeId, from p: IntPoint, to: Direction, isTest: Bool = true, isSeed: Bool = false) -> Int {
        let q = p.neighbor(to)
        if !rect.contains(q) {
            extend(toInclude: q)
        }
        
        guard let i = index(q) else { return 0 }
        
        var bondCount = 0
        beadTypeId[i] = typeId
        let bond : Bond = isTest ? .testBond : .bond
        let opp = to.opposite
        for d in Direction.all where d != opp {
            if let j = index(q.neighbor(d)) {
                if let v = beadTypeId[j], let r = rule, r.attraction(between: typeId, and: v) {
                    setBond(at: q, to: d, value: bond)
                    bondCount += 1
                } else {
                    setBond(at: q, to: d, value: nil)
                }
            }
        }
        
        setBond(at: p, to: to, value: isTest ? (isSeed ? .testSeed : .testPrimary) : (isSeed ? .seed : .primary))
        setOutEdge(at: p, to: to)
        
        return bondCount
    }
    
    @inline(__always) func addBead(typeId: BeadTypeId, at p: IntPoint, bond: Bond) -> Int {
        if !rect.contains(p) {
            extend(toInclude: p)
        }
        
        guard let i = index(p) else { return 0 }
        
        var bondCount = 0
        beadTypeId[i] = typeId
        for d in Direction.all {
            if let j = index(p.neighbor(d)) {
                if let v = beadTypeId[j], let r = rule, r.attraction(between: typeId, and: v) {
                    setBond(at: p, to: d, value: bond)
                    bondCount += 1
                } else {
                    setBond(at: p, to: d, value: nil)
                }
            }
        }
        return bondCount
    }
    


    @inline(__always) func removeBead(at p: IntPoint) {
        guard let i = index(p) else { return }
        
        beadTypeId[i] = nil
        for d in Direction.all {
            setBond(at: p, to: d, value: nil)
        }
        setOutEdge(at: p, to: nil)
    }
    
    @inline(__always) func bondFor(test: Bool = true, seed: Bool = false) -> Bond {
        return test ? .testBond : .bond
    }
    
    @inline(__always) func primaryBondFor(test: Bool = true, seed: Bool = false) -> Bond {
        return test ? (seed ? .testSeed : .testPrimary) : (seed ? .seed : .primary)
    }
    
    func draw(path: Path, beadTypeStream: (Int) -> BeadTypeId, isTesting: Bool = false, isSeed: Bool = false, from p: IntPoint, in range: Range<Int>? = nil) -> (nbBonds: Int, endPoint: IntPoint) {

        var q = p
        var bq = beadTypeId(at: q)
        var idx = beadIndex(at: p) ?? 0
        var bondCount = 0
        let bond : Bond = bondFor(test: isTesting, seed: isSeed)
        let primary : Bond = primaryBondFor(test: isTesting, seed: isSeed)

        let range = range ?? 0..<path.count
        for i in range {
            let d = path[i]
            let v = q.neighbor(d)
            let bi = beadTypeStream(i)
            bondCount += addBead(typeId: bi, at: v, bond: bond)
            if let b = bq {
                if let r = rule, r.attraction(between: b, and: bi) {
                    bondCount -= 1
                }
                setBond(at: q, to: d, value: primary)
                setOutEdge(at: q, to: d)
            }
            idx += 1
            setBeadIndex(at: v, value: idx)
            q = v
            bq = bi
        }
        return (bondCount, q)
    }

    func draw(conformation: Conformation, isTesting: Bool = false, isSeed: Bool = false, from p: IntPoint = IntPoint.zero) -> (nbBonds: Int, endPoint: IntPoint) {
        var q = p
        var bq = beadTypeId(at: q)
        var idx = beadIndex(at: p) ?? 0
        var bondCount = 0
        let bond : Bond = bondFor(test: isTesting, seed: isSeed)
        let primary : Bond = primaryBondFor(test: isTesting, seed: isSeed)

        for (path, beads) in conformation {
            let expanded_path = path.expand()
            guard expanded_path.count == beads.count else {
                print("In conformation \(conformation), the lengths of path \(path) and beads \(beads) do not match")
                return (bondCount, q)
            }
            for i in 0..<expanded_path.count {
                let d = expanded_path[i]
                let v = q.neighbor(d)
                let bi = beads[i]
            bondCount += addBead(typeId: bi, at: v, bond: bond)
                if let b = bq {
                    if let r = rule, r.attraction(between: b, and: bi) {
                        bondCount -= 1
                    }
                    setBond(at: q, to: d, value: primary)
                    setOutEdge(at: q, to: d)
                }
                idx += 1
                setBeadIndex(at: v, value: idx)
                q = v
                bq = bi
            }
        }
        return (bondCount, q)
    }


    
    func erase(path: Path, from p: IntPoint, shortenBy start: Int = 0) {
        var q = p
        for i in start..<path.count-start {
            q = q.neighbor(path[i])
            removeBead(at: q)
        }
    }
    
    func beadTypeStream() -> [BeadTypeId] {
        guard pathStartingPoints.count > 0 else { return [] }
        var s : [BeadTypeId] = []
        synchronize {
            var p = pathStartingPoints[0]
            s = [beadTypeId(at:p)!]
            while let d = outEdge(at: p) {
                p = p.neighbor(d)
                s.append(beadTypeId(at: p)!)
            }
        }
        return s
    }

    func beadPath() -> Path {
        guard pathStartingPoints.count > 0 else { return [] }
        var p = pathStartingPoints[0]
        var path : Path = []
        while let d = outEdge(at: p) {
            path.append(d)
            p = p.neighbor(d)
        }
        return path
    }
    
    // EXTENSION
    
    func extend(toInclude p:IntPoint) {
        synchronizeIfNotAlreadySynchronized {
            let oldBeadIndex = beadIndex
            let oldOutEdge = outEdge
            let oldBeadTypeId = beadTypeId
            let oldBond = bond
            let oldInfo = info
            let oldRect = rect
            
            // Extend and allocate
            rect.extend(toInclude: p)
            allocArrays()
            
            print("New rect: \(rect)")
            
            // Copy old content
            for y in 0..<oldRect.height {
                var oldIndex = y * oldRect.width
                var newIndex = (y + oldRect.y - rect.y) * rect.width + oldRect.x - rect.x
                for _ in 0..<oldRect.width {
                    beadIndex[newIndex] = oldBeadIndex[oldIndex]
                    outEdge[newIndex] = oldOutEdge[oldIndex]
                    beadTypeId[newIndex] = oldBeadTypeId[oldIndex]
                    for z in 0..<Direction.halfCount {
                        bond[Direction.halfCount * newIndex + z] = oldBond[Direction.halfCount * oldIndex + z]
                    }
                    info[newIndex] = oldInfo[oldIndex]
                    
                    oldIndex += 1
                    newIndex += 1
                }
            }
        }
    }
}


extension TriangularGrid {
    
    func pathBounds()-> CGRect {
        guard pathStartingPoints.count > 0 else { return CGRect.zero }
        var p = pathStartingPoints[0]
        let q = p.point
        var xmin = q.x
        var xmax = q.x
        var ymin = q.y
        var ymax = q.y
        
        while let d = outEdge(at: p) {
            p = p.neighbor(d)
            let q = p.point
            if q.x < xmin { xmin = q.x }
            if q.x > xmax { xmax = q.x }
            if q.y < ymin { ymin = q.y }
            if q.y > ymax { ymax = q.y }
        }
        
        return CGRect(x: xmin, y: ymin, width: xmax-xmin, height: ymax-ymin)
    }
    
    func expand(rect: CGRect, byX x: Int, y: Int) -> CGRect {
        let p = IntPoint(closestTo: CGPoint(x: rect.minX, y: rect.minY))
        let q = IntPoint(closestTo: CGPoint(x: rect.maxX, y: rect.maxY))
        
        let pp = IntPoint(x: p.x - x, y: p.y - y).point
        let qq = IntPoint(x: q.x + x, y: q.y + y).point
        
        return CGRect(x: pp.x, y: pp.y, width: qq.x - pp.x, height: qq.y - pp.y)
    }
    
}


// Compute bond direction for folding with delay 1
extension TriangularGrid {
    
    static func boundingRect(for path: Path) -> CGRect {
        var (minX, maxX, minY, maxY) : (CGFloat,CGFloat,CGFloat,CGFloat) = (0, 0, 0, 0)
        var q = IntPoint.zero
        for d in path {
            q = q.neighbor(d)
            let p = q.point
            if p.x < minX { minX = p.x }
            if p.x > maxX { maxX = p.x }
            if p.y < minY { minY = p.y }
            if p.y > maxY { maxY = p.y }
        }
        return CGRect(x: minX, y: minY, width: maxX-minX, height: maxY-minY)
    }
    
    func occupiedNeighborDirection(from p: IntPoint) -> Direction {
        for d in Direction.all {
            let v = p.neighbor(d)
            if  isOccupied(at: v) {
                return d
            }
        }
        return Direction(rawValue: 0)!
    }
}
