//
//  TriangularGrid.swift
//  Scary Pacman
//
//  Created by Nikao on 19/9/17.
//  Copyright © 2017 Algoseum. All rights reserved.
//


// Séparer en deux class: Triangular grid (contient échelle + routine de base: récupérer du point de cliquage etc.)+ pathconstructor avec une classe pour chaque ensemble d'échelle + peut-être une classe pour mémoriser la forme construire pour pouvoir la passer à différentes échelles.

// Réparer cellNeighbor -> trouver un fix pour les échelles x.5 et l'échelle 3.0, ne pas chercher des fonctions trop générale si pas utile pour échelle autres que 3.0 (cf par exemple la nécessité de regarder les arêtes -> pourquoi ne pas stocker les anomalies au centre de la cellule? est-ce vraiment utile?)

import Foundation
import QuartzCore

enum GridInfo : Int, CustomStringConvertible {
    case none
    var description: String {
        switch self {
        case .none: return "None"
        }
    }
}

typealias BeadTypeId = Int

protocol TriangularGridProtocol : class {
    
    func synchronize(closure: () -> ())
    
    func synchronizeIfNotAlreadySynchronized(closure: () -> ())
        
    var minX: Int { get }
    var maxX: Int { get }
    var minY: Int { get }
    var maxY: Int { get }

    var pathStartingPoints: [ IntPoint ] { get set }
    
    var beadTypeSet : BeadTypeSet? { get set }
    var rule: OritatamiRule? { get set }
    
    func clear()
    
    func isOccupied(at p:IntPoint) -> Bool
    
    func beadTypeId(at p:IntPoint) -> BeadTypeId?
    func setBeadTypeId(at p:IntPoint, value: BeadTypeId?)
    
    func beadIndex(at p:IntPoint) -> Int?
    func setBeadIndex(at p:IntPoint, value: Int?)

    func bond(at p: IntPoint, direction d: Direction) -> Bond?
    func setBond(at p: IntPoint, to d: Direction, value: Bond?)
    
    func outEdge(at p:IntPoint) -> Direction?
    func setOutEdge(at p: IntPoint, to d: Direction?)
    
    func addBead(typeId: BeadTypeId, from p: IntPoint, to: Direction, isTest: Bool, isSeed: Bool) -> Int
    func addBead(typeId: BeadTypeId, at p: IntPoint, bond: Bond) -> Int
    func removeBead(at p: IntPoint)
    
    func bondFor(test: Bool, seed: Bool) -> Bond
    func primaryBondFor(test: Bool, seed: Bool) -> Bond
    
    func draw(path: Path, beadTypeStream: (Int) -> BeadTypeId, isTesting: Bool, isSeed: Bool, from p: IntPoint, in range: Range<Int>?) -> (nbBonds: Int, endPoint: IntPoint)
    func draw(conformation: Conformation, isTesting: Bool, isSeed: Bool, from p: IntPoint) -> (nbBonds: Int, endPoint: IntPoint)

    
    func erase(path: Path, from p: IntPoint, shortenBy start: Int)
    
    func beadTypeStream() -> [BeadTypeId]

    func beadPath() -> Path
    
    func pathBounds()-> CGRect
    
    func expand(rect: CGRect, byX x: Int, y: Int) -> CGRect

    func occupiedNeighborDirection(from p: IntPoint) -> Direction

}

