//
//  PacmanView.swift
//  Scary Pacman
//
//  Created by Nikao on 19/9/17.
//  Copyright © 2017 Algoseum. All rights reserved.
//

import UIKit

@IBDesignable
class TriangularGridView: UIView {

    var controller : TriangularGridViewController?
    
    override var canBecomeFirstResponder: Bool {
        return true
    }
    
    weak var grid: TriangularGridProtocol? = nil { didSet { setNeedsDisplay() } }
    
    var mapOrigin: CGPoint = CGPoint.zero  { didSet { setNeedsDisplay() } }

    var mapCenter: CGPoint = CGPoint.zero  { didSet { setNeedsDisplay() } }

    @IBInspectable var scale: CGFloat = Defaults.scale  {
        didSet { scale = max(Defaults.minScale, scale); setNeedsDisplay() }
    }
    var gridColor: __Color = Defaults.gridColor
    var upscaledGridColor: __Color = Defaults.upscaledGridColor
    var gridLineWidth: CGFloat = Defaults.gridLineWidth
    var upscaledGridLineWidth: CGFloat = Defaults.upscaledGridLineWidth
    var primaryBondColorUpfront: __Color = Defaults.primaryBondColorUpfront
    var primaryBondColorUpfrontPDF: __Color = Defaults.primaryBondColorUpfrontPDF
    var primaryBondColorBehind: __Color = Defaults.primaryBondColorBehind
    var primaryBondLineWidthUpfront: CGFloat = Defaults.primaryBondLineWidthUpfront
    var primaryBondLineWidthBehind: CGFloat = Defaults.primaryBondLineWidthBehind
    var primaryTestBondLineWidthUpfront: CGFloat = Defaults.primaryTestBondLineWidthUpfront
    @IBInspectable var primaryTestBondLineWidthBehind: CGFloat = Defaults.primaryTestBondLineWidthBehind
    @IBInspectable var primaryTestBondColor: __Color = Defaults.primaryTestBondColor
    @IBInspectable var secondaryBondColor: __Color = Defaults.secondaryBondColor
    @IBInspectable var secondaryBondLineWidth: CGFloat = Defaults.secondaryBondLineWidth
    @IBInspectable var secondaryTestBondLineWidth: CGFloat = Defaults.secondaryTestBondLineWidth
    @IBInspectable var secondaryBondLineDash: [CGFloat] = Defaults.secondaryBondLineDash
    static var beadRadius: CGFloat = Defaults.beadRadius
    @IBInspectable var beadInnerRadius: CGFloat = Defaults.beadInnerRadius
    @IBInspectable var beadColors: [__Color] = Defaults.beadColors
    static var beadLineWidth: CGFloat = Defaults.beadLineWidth
    static var beadLineColor: __Color = Defaults.beadLineColor
    @IBInspectable var blockedBeadColor: __Color = Defaults.blockedBeadColor
    @IBInspectable var anomalyBeadColor: __Color = Defaults.anomalyBeadColor
    @IBInspectable var arrowProbability: Double = Defaults.arrowProbability
    @IBInspectable var hueStep: CGFloat = Defaults.hueStep
    @IBInspectable var hexagonMinBrightness: CGFloat = Defaults.hexagonMinBrightness
    @IBInspectable var hexagonMaxBrightness: CGFloat = Defaults.hexagonMaxBrightness
    
    var displayPath : Bool = true

    override func draw(_ rect: CGRect) {
        guard !isHidden else { return }
        let ctx = __GraphicsGetCurrentContext()!
        
        guard let g = grid else { return }

//        if (scale >= 10) { drawGrid(rect, in: ctx) }

        g.synchronize {
            if scale > Defaults.minDrawBeadScale {
                drawBonds(g, rect: rect, in: ctx)
                drawBeads(g, rect: rect, in: ctx, with: TriangularGridView.drawBead(_:at:in:))
            } else {
                if let g = g as? TriangularGridDictionary {
                    drawCompressedPath(g, rect: rect, in: ctx)
                } else {
                    drawPrimaryStructure(g, rect: rect, in: ctx)
                }
            }
            
            if Defaults.show_debug_grid_dictionnary {
                draw_debug_grid_dictionnary(in: ctx)
            }
        }
    }
    
    private func draw_debug_grid_dictionnary(in ctx: CGContext) {
        if let gd = grid as? TriangularGridDictionary {
            ctx.saveGState()
            
            // scale properly
            ctx.scaleBy(x: scale, y: scale)
            ctx.translateBy(x: -mapCenter.x + bounds.width/2/scale,
                            y: -mapCenter.y + bounds.height/2/scale)

            // linewidth
            ctx.setLineWidth(secondaryBondLineWidth)

            // Draw dictionaries positions
            let (db,ds) = (gd.dict_bitsize,
                           gd.dict_size-1)
            for p in gd.dict_of_dict.keys {
                let _p = IntPoint(x: p.x<<db, y: p.y<<db)
                ctx.move(to: _p.point)
                for (x,y) in [(0,ds), (ds,ds), (ds,0)] {
                    ctx.addLine(to: IntPoint(x: _p.x+x, y: _p.y+y).point)
                }
                ctx.closePath()
            }
            ctx.setStrokeColor(UIColor.green.cgColor)
            ctx.strokePath()
            
            // Draw the cache position
            ctx.setStrokeColor(UIColor.red.cgColor)
            let co = gd.cache_origin
            ctx.move(to: co.point)
            let cs = gd.cache_size
            for (x,y) in [(cs-1,0), (cs-1,cs-1), (0,cs-1)] {
                ctx.addLine(to: (co + IntPoint(x: x, y: y)).point)
            }
            ctx.closePath()
            
            let ccs = gd.cache_cell_size
            let cc = co+IntPoint(x: ccs, y: ccs)
            ctx.move(to: cc.point)
            let ccs2 = 2*ccs-1
            for (x,y) in [(ccs2,0), (ccs2,ccs2), (0,ccs2)] {
                ctx.addLine(to: (cc + IntPoint(x: x, y: y)).point)
            }
            ctx.closePath()
            
            // Draw the committed points
            for p in gd.committed_points {
                ctx.addEllipse(in: CGRect.init(circleAt: p.point, radius: 1.3*Self.beadRadius))
            }

            ctx.strokePath()
            ctx.restoreGState()
        }
    }
    
    
    private static func drawBead(_ grid : TriangularGridProtocol, at p: IntPoint, in ctx: __CGContext) {
        guard let bts = grid.beadTypeSet else { print("arg1"); return }
        guard let bti = grid.beadTypeId(at: p) else { print("arg2"); return }
        let q = p.point
        let br = beadRadius
        let bt = bts.beadType[bti]
        let c = bt.color.cgColor
        
        ctx.saveGState()

        ctx.setFillColor(c)
        ctx.setStrokeColor(beadLineColor.cgColor)
        ctx.setLineWidth(beadLineWidth)

        ctx.translateBy(x: q.x, y: q.y)
        let rect = CGRect(x: -br, y: -br, width: 2*br, height:2*br)

        ctx.addEllipse(in: rect)
        ctx.drawPath(using: .fillStroke)

        //        print(__Font.familyNames)
        let fontName = Defaults.patternFontName
        var fontSize = Defaults.patternFontSize
        var font = __Font(name: fontName, size: fontSize) ?? __Font.boldSystemFont(ofSize: fontSize)
        __GraphicsPushContext(ctx)
        var name = NSAttributedString(string: bt.name, attributes: [.font: font, .foregroundColor: __Color.black])
        var textSize = name.size()
        if textSize.width > 1.8*br {
            fontSize = Defaults.patternFontSize * 1.8 * br / textSize.width

            let category = bts.category[bt.categoryId].name
            let number = "\(bt.number)"
            var _category = NSAttributedString(string: category, attributes: [.font: font, .foregroundColor: __Color.black])
            var _number = NSAttributedString(string: number, attributes: [.font: font, .foregroundColor: __Color.black])
            let cs = _category.size()
            let ns = _number.size()
            let rc = sqrt(cs.width*cs.width/4 + cs.height*cs.height)
            let rn = sqrt(ns.width*ns.width/4 + ns.height*ns.height)
            let mx_r = max(rc, rn)
            let fontSize2 = Defaults.patternFontSize * min(1, 1.1*br / mx_r)
            if fontSize2 > fontSize {
                font = __Font(name: fontName, size: fontSize2) ?? __Font.boldSystemFont(ofSize: fontSize2)
                _category = NSAttributedString(string: category, attributes: [.font: font, .foregroundColor: __Color.black])
                _number = NSAttributedString(string: number, attributes: [.font: font, .foregroundColor: __Color.black])
                let categorySize = _category.size()
                let numberSize = _number.size()
                _category.draw(at: CGPoint(x:-categorySize.width/2,y:-categorySize.height))
                _number.draw(at: CGPoint(x:-numberSize.width/2,y:0))
                __GraphicsPopContext()
                ctx.restoreGState()
                return
            }
            font = __Font(name: fontName, size: fontSize) ?? __Font.boldSystemFont(ofSize: fontSize)
            name = NSAttributedString(string: bt.name, attributes: [.font: font, .foregroundColor: __Color.black])
            textSize = name.size()
        }
        name.draw(at: CGPoint(x:-textSize.width/2,y:-textSize.height/2))
        __GraphicsPopContext()
        
        ctx.restoreGState()
        
//        print("Bead \(bt) drawn at \(p)")
    }

    
    
    func addEdge(from p: CGPoint, to q: CGPoint, in ctx: __CGContext) {
        if displayPath && Darwin.drand48() < arrowProbability {
            ctx.makeArrow(from: p, to: q)
        } else {
            ctx.move(to: p)
            ctx.addLine(to: q)
        }
    }
    
    func centerOrigin() {
        print("center")
        mapOrigin = CGPoint(x: bounds.midX, y:bounds.midY)
        setNeedsDisplay()
    }
    
    func drawBonds(_ grid: TriangularGridProtocol, rect: CGRect, in ctx: __CGContext) {
        // Paths for every type of bonds
        var paths : [ CGMutablePath ] = []
        for _ in 0..<Bond.count {
            paths.append(CGMutablePath())
        }
        
        ctx.saveGState()
        
        ctx.clip(to: rect)
        
//        ctx.translateBy(x: mapOrigin.x, y: mapOrigin.y)
//        ctx.scaleBy(x: scale, y: scale)
        ctx.scaleBy(x: scale, y: scale)
        ctx.translateBy(x: -mapCenter.x + bounds.width/2/scale,
                        y: -mapCenter.y + bounds.height/2/scale)

//        let normRect = rect.unMoveScale(mapOrigin, scale).insetBy(dx: -1, dy: -1)
        let normRect = ((rect - bounds.center)/scale + mapCenter).insetBy(dx: -1, dy: -1)
        //        let iRect = IntRect(englobing: rect.unshiftScale(mapOrigin, scale))
        let pMax = IntPoint(closestTo: CGPoint(x:normRect.maxX, y: normRect.maxY))
        let pMin = IntPoint(closestTo: CGPoint(x:normRect.minX, y: normRect.minY))

        // Bonds
        let drawDirections: [Direction] = [ .E, .SE, .SW, ]
        let yMin = min(grid.maxY,max(grid.minY,pMin.y))
        let yMax = min(grid.maxY,max(grid.minY,pMax.y))

        for y in yMin...yMax {
            let yf = IntPoint(x:0, y: y).point.y
            let xMin = min(grid.maxX, max(grid.minX, IntPoint(closestTo: CGPoint(x:normRect.minX, y: yf)).x))
            let xMax = max(grid.minX, min(grid.maxX, IntPoint(closestTo: CGPoint(x:normRect.maxX, y: yf)).x))
            for x in xMin...xMax {
                let p = IntPoint(x:x, y:y)
                guard normRect.contains(p.point) else { continue }
                for d in drawDirections {
                    if let b = grid.bond(at: p, direction: d) {
                        let v = p.neighbor(d)
                        paths[b.rawValue].addLines(between: [p.point, v.point])
                    }
                }
            }
        }
        
        ctx.setLineJoin(CGLineJoin.miter)
        ctx.setLineCap(CGLineCap.butt)
        for b in Bond.all {
            let bondStyle = Defaults.bondStyle(b)
            ctx.setLineWidth(bondStyle.lineWidth)
            ctx.setLineDash(phase: 0, lengths: bondStyle.dashPattern)
            ctx.setStrokeColor(bondStyle.strokeColor.cgColor)

            let p = paths[b.rawValue]
            ctx.addPath(p)
            ctx.strokePath()
        }
        
        ctx.restoreGState()
    }
    
    func drawCompressedPath(_ grid: TriangularGridDictionary, rect: CGRect, in ctx: __CGContext, saveAsPDF: Bool = false) {
        
        ctx.saveGState()
        
        if !saveAsPDF {
            ctx.scaleBy(x: scale, y: scale)
            ctx.translateBy(x: -mapCenter.x + bounds.width/2/scale,
                            y: -mapCenter.y + bounds.height/2/scale)
            let line_width : CGFloat = max(
                Defaults.min_compressed_path_line_width / scale,
                Defaults.bondStyle(.primary).lineWidth)
            ctx.setLineWidth(line_width)
        } else {
            ctx.setLineWidth(Defaults.bondStyle(.primary).lineWidth)
        }

        guard grid.pathStartingPoints.count > 0 else { return }

        ctx.setLineCap(.round)
        ctx.setLineJoin(.round)
        
        grid.compressed_path.stroke(from: grid.pathStartingPoints[0], in: ctx)

        ctx.restoreGState()
    }
    
    func drawPrimaryStructure(_ grid: TriangularGridProtocol, rect: CGRect, in ctx: __CGContext, saveAsPDF: Bool = false) {
        // Paths for every type of bonds

        ctx.saveGState()
        
//        ctx.clip(to: rect)
        
//        ctx.translateBy(x: mapOrigin.x, y: mapOrigin.y)
//        ctx.scaleBy(x: scale, y: scale)
       
        if !saveAsPDF {
            ctx.scaleBy(x: scale, y: scale)
            ctx.translateBy(x: -mapCenter.x + bounds.width/2/scale,
                            y: -mapCenter.y + bounds.height/2/scale)
        }
        
//        print(grid.pathStartingPoints)
        guard grid.pathStartingPoints.count > 0 else { return }
        
        ctx.setLineWidth(Defaults.bondStyle(.primary).lineWidth)
        ctx.setLineCap(.butt)
        ctx.setLineJoin(.round)
        
        for start in grid.pathStartingPoints {
            var p = start
            var prev = start
            
            var pP = p.point
            var prevP = prev.point
            
            var notDone = true
            repeat {
                guard let bts = grid.beadTypeSet else { print("arg1"); return }
                guard let bti = grid.beadTypeId(at: p) else { print("arg2"); return }
                let bt = bts.beadType[bti]
                let c = bt.color.cgColor
                
                var next : IntPoint

                if let d = grid.outEdge(at: p) {
                    next = p.neighbor(d)
                } else {
                    next = p
                    notDone = false
                }
                let nextP = next.point
                
                ctx.setStrokeColor(c)
                ctx.addLines(between: [(prevP+pP)/2, pP, (pP+nextP)/2])
                ctx.strokePath()
                
                prev = p
                prevP = pP
                
                p = next
                pP = nextP
                
            } while notDone
        }
        
        ctx.restoreGState()
    }
    

    func drawBeads(_ grid: TriangularGridProtocol, rect:CGRect, in ctx: __CGContext, with drawBeadMethod: (TriangularGridProtocol, IntPoint, __CGContext) -> () = TriangularGridView.drawBead(_:at:in:), drawGrid: Bool = true, drawBeads: Bool = true) {
        ctx.saveGState()
        
        ctx.clip(to: rect)


        //        ctx.translateBy(x: mapOrigin.x, y: mapOrigin.y)
        //        ctx.scaleBy(x: scale, y: scale)
        ctx.scaleBy(x: scale, y: scale)
        ctx.translateBy(x: -mapCenter.x + bounds.width/2/scale,
                        y: -mapCenter.y + bounds.height/2/scale)
        
        let drawGrid = drawGrid && (pdf_force_draw_grid  || (scale >= Defaults.minDrawGridScale))
        
//        ctx.setFillColor(__Color.red400.cgColor)
//        ctx.fill(CGRect(x:mapCenter.x, y:mapCenter.y, width:5, height:5))
//        return
        
//        let normRect = rect.unMoveScale(mapOrigin, scale).insetBy(dx: -1, dy: -1)
        let normRect = ((rect - bounds.center)/scale + mapCenter).insetBy(dx: -1, dy: -1)
        //        let iRect = IntRect(englobing: rect.unshiftScale(mapOrigin, scale))
        let pMax = IntPoint(closestTo: CGPoint(x:normRect.maxX, y: normRect.maxY))
        let pMin = IntPoint(closestTo: CGPoint(x:normRect.minX, y: normRect.minY))
        
        let minY = min(grid.maxY, max(grid.minY,pMin.y))
        let maxY = max(grid.minY,min(grid.maxY,pMax.y))
        
        for y in minY...maxY {
            let yf = IntPoint(x:0, y: y).point.y
            let minX = min(grid.maxX, max(grid.minX, IntPoint(closestTo: CGPoint(x:normRect.minX, y: yf)).x))
            let maxX = min(grid.maxX, max(grid.minX, IntPoint(closestTo: CGPoint(x:normRect.maxX, y: yf)).x))
            for x in minX...maxX {
                let p = IntPoint(x:x, y:y)
                guard normRect.contains(p.point) else { continue }
                if grid.isOccupied(at: p) {
                    if drawBeads { drawBeadMethod(grid, p, ctx) }
                } else {
                    if !drawGrid { continue }
                    ctx.setFillColor(Defaults.gridColor.cgColor)
//                    ctx.setFillColor(__Color(hue:CGFloat.random(in:0...1), saturation: 1, brightness: 1, alpha:1).cgColor)
                    let r = Defaults.unoccupiedBeadRadius // * CGFloat.random(in:0...10)
                    ctx.fillEllipse(in: CGRect(x: p.point.x - r, y: p.point.y - r, width: 2*r, height: 2*r))
                }
            }
        }
        
        ctx.setLineJoin(CGLineJoin.miter)
        ctx.setLineCap(CGLineCap.round)
        ctx.setStrokeColor(TriangularGridView.beadLineColor.cgColor)
        ctx.setLineWidth(TriangularGridView.beadLineWidth)
        
        ctx.strokePath()
        
        ctx.restoreGState()
    }

    
//    func drawGrid(_ rect: CGRect, in ctx: __CGContext) {
//        ctx.saveGState()
//        ctx.clip(to: rect)
//        ctx.translateBy(x: mapOrigin.x, y: mapOrigin.y)
//        ctx.scaleBy(x: scale, y: scale)
//
//        let iRect = IntRect(englobing: rect.unshiftScale(mapOrigin, scale))
//
////        ctx.stroke(iRect.cgRect.shiftScale(mapOrigin, scale))
//
//        let h = CGFloat(iRect.height)
//        let hs = h * IntPoint.sqrt32
//        let w = CGFloat(iRect.width)
//
////        let o = iRect.origin.point
////        let p = o.shiftScale(mapOrigin, scale)
//        let p = iRect.origin.point
//        for i in 0...iRect.width+iRect.height {
//            let i_f = CGFloat(i)
//            ctx.move(to: CGPoint(x: p.x + i_f, y: p.y))
//            ctx.addLine(to: CGPoint(x: p.x + i_f - h/2, y: p.y + hs))
//
//            let j_f = CGFloat(i-iRect.height)
//            ctx.move(to: CGPoint(x: p.x + j_f, y: p.y))
//            ctx.addLine(to: CGPoint(x: p.x + j_f + h/2, y: p.y + hs))
//        }
//
//        for i in 0...iRect.height {
//            let y = CGFloat(i) * IntPoint.sqrt32
//            ctx.move(to: CGPoint(x: p.x, y: p.y + y ))
//            ctx.addLine(to: CGPoint(x: p.x + w, y: p.y + y))
//        }
//
//        ctx.setStrokeColor(gridColor.cgColor)
//        ctx.setLineWidth(gridLineWidth)
//
//        ctx.strokePath()
//
//        ctx.restoreGState()
//    }
    
    func point(at p: CGPoint) -> CGPoint {
//        return (p - mapOrigin)/scale
        return mapCenter + (p - bounds.center)/scale
    }
    
    func intPoint(at p: CGPoint) -> IntPoint {
        let q = point(at: p)
        return IntPoint(closestTo: q)
    }
    
    func center(at p: CGPoint) {
//        mapOrigin = CGPoint(x: bounds.midX, y: bounds.midY) - p * scale
        mapCenter = p
        setNeedsDisplay()
    }
    
    private var mapOriginTimer: Timer? = nil
    private var mapOriginOrigin : CGPoint = CGPoint.zero
    private var mapOriginDestination : CGPoint = CGPoint.zero
    private var mapOriginStartingTime: TimeInterval = 0
    private let mapOriginTimingFunction = { (t: TimeInterval) -> CGFloat in
        let u = 1-CGFloat(t)
        return 1-u*u*u //sin(CGFloat.pi/2 * CGFloat(t))
    }
    
    func moveMapOrigin(to p: CGPoint) {
        if let t = mapOriginTimer, t.isValid {
            t.invalidate()
            mapOriginTimer = nil
        }
        mapOriginOrigin = mapOrigin
        mapOriginDestination = p
        mapOriginStartingTime = CACurrentMediaTime()
        mapOriginTimer = Timer.scheduledTimer(withTimeInterval: 1/30, repeats: true) {
            [weak self] (timer: Timer) in
            guard let me = self else { timer.invalidate(); return }
            let now = CACurrentMediaTime()
            let t = min(1, (now - me.mapOriginStartingTime) / Defaults.mapOriginMoveLength)
            if t >= 1 {
                me.mapOrigin = me.mapOriginDestination
                timer.invalidate()
                me.mapOriginTimer = nil
                return
            }
            let x = me.mapOriginTimingFunction(t)
            me.mapOrigin = me.mapOriginOrigin * (1-x) + x * me.mapOriginDestination
            me.setNeedsDisplay()
        }
    }

    private var mapCenterTimer: Timer? = nil
    private var mapCenterStart : CGPoint = CGPoint.zero
    private var mapCenterEnd : CGPoint = CGPoint.zero
    private var mapCenterStartingTime: TimeInterval = 0
    private let mapCenterTimingFunction = { (t: TimeInterval) -> CGFloat in
        let u = 1-CGFloat(t)
        return 1-u*u*u //sin(CGFloat.pi/2 * CGFloat(t))
    }
    
    func moveMapCenter(to p: CGPoint) {
        if let t = mapCenterTimer, t.isValid {
            t.invalidate()
            mapCenterTimer = nil
        }
        mapCenterStart = mapCenter
        mapCenterEnd = p
        mapCenterStartingTime = CACurrentMediaTime()
        mapCenterTimer = Timer.scheduledTimer(withTimeInterval: 1/30, repeats: true) {
            [weak self] (timer: Timer) in
            guard let me = self else { timer.invalidate(); return }
            let now = CACurrentMediaTime()
            let t = min(1, (now - me.mapCenterStartingTime) / Defaults.mapOriginMoveLength)
            if t >= 1 {
                me.mapCenter = me.mapCenterEnd
                timer.invalidate()
                me.mapCenterTimer = nil
                return
            }
            let x = me.mapCenterTimingFunction(t)
            me.mapCenter = me.mapCenterStart * (1-x) + x * me.mapCenterEnd
            me.setNeedsDisplay()
        }
    }

    
//    func centerIfNeeded(at p: IntPoint, animated: Bool = true) {
//        let q = point(for: p)
//        let b = bounds
//        var newO = mapOrigin
//        var changed = false
//        if q.x < b.minX + 0.2*b.width || q.x > b.maxX - 0.2*b.width {
//            newO.x = b.midX - p.point.x * scale
//            changed = true
//        }
//        if q.y < b.minY + 0.2*b.height || q.y > b.maxY - 0.2*b.height {
//            newO.y = b.midY - p.point.y * scale
//            changed = true
//        }
//        if changed {
//            if !animated {
//            mapOrigin = newO
//            setNeedsDisplay()
//            } else {
//                moveMapOrigin(to: newO)
//            }
//        }
//    }

    func centerIfNeeded(at p: IntPoint, animated: Bool = true) {
        let q = point(for: p)
        let b = bounds
        var centerAtP = false
        if q.x < b.minX + 0.2*b.width || q.x > b.maxX - 0.2*b.width {
            centerAtP = true
        }
        if q.y < b.minY + 0.2*b.height || q.y > b.maxY - 0.2*b.height {
            centerAtP = true
        }
        if centerAtP {
            if !animated {
                center(at: p.point)
            } else {
                moveMapCenter(to: p.point)
            }
        }
    }
    
    
    func point(for p:IntPoint) -> CGPoint {
//        return p.point*scale + mapOrigin
        return bounds.center + scale * (p.point - mapCenter)
    }
    
    func scaledRect(_ rect:CGRect) -> CGRect {
//        return rect * scale + mapOrigin
        return (rect - mapCenter) * scale + bounds.center
    }

    func setNeedsDisplay(inUnscaled rect:CGRect) {
        setNeedsDisplay(scaledRect(rect))
    }

    var pdf_force_draw_grid = false
}

// PDF output

extension TriangularGridView {
    func saveToPDFDocumentFile(_ filename: String) -> Bool {
        let fileManager = FileManager.default
        do {
            let documentDirectory = try fileManager.url(for: .documentDirectory, in: .userDomainMask, appropriateFor:nil, create:false)
            let fileURL = documentDirectory.appendingPathComponent(filename)
            //let ctx = CGPDFContextCreateWithURL(fileURL, self.bounds, nil)
            __GraphicsBeginPDFContextToFile(fileURL.path, self.bounds, nil)
            __GraphicsBeginPDFPage()
            self.draw(self.bounds)
            __GraphicsEndPDFContext()
        } catch {
            Swift.print(error)
            return false
        }
        return true
    }

    func saveToPathPDFDocumentFile(_ filename: String, compressedIfPossible: Bool = false) -> Bool {
        print("Saving...")
        let fileManager = FileManager.default
        do {
            guard let g = grid else { return false }
            var p_bounds: CGRect
            if compressedIfPossible, let g = g as? TriangularGridDictionary {
                p_bounds = g.compressed_path.bounds(from: g.pathStartingPoints[0])
            } else {
                p_bounds = g.pathBounds()
            }
            let rect = g.expand(rect: p_bounds, byX: 2, y: 2)
            let documentDirectory = try fileManager.url(for: .documentDirectory, in: .userDomainMask, appropriateFor:nil, create:false)
            let fileURL = documentDirectory.appendingPathComponent(filename)
            print("Saving to \(fileURL))")
            //let ctx = CGPDFContextCreateWithURL(fileURL, self.bounds, nil)
            print(rect)
            __GraphicsBeginPDFContextToFile(fileURL.path, CGRect(x:0, y:0, width: rect.width, height: rect.height), nil)
            __GraphicsBeginPDFPage()
            let ctx = __GraphicsGetCurrentContext()!
            ctx.translateBy(x: -rect.minX, y: -rect.minY)
            if compressedIfPossible, let g = g as? TriangularGridDictionary {
                drawCompressedPath(g, rect: p_bounds, in: ctx, saveAsPDF: true)
            } else {
                drawPrimaryStructure(g, rect: p_bounds, in: ctx, saveAsPDF: true)
            }
            __GraphicsEndPDFContext()
        } catch {
            Swift.print(error)
            return false
        }
        return false
    }

    func saveViewToPDFDocumentFile(_ filename: String) -> Bool {
        print("Saving...")
        let fileManager = FileManager.default
        do {
            let target_scale = 2.5*Defaults.minDrawGridScale
            let scale_change = target_scale / scale
            let rect = bounds * scale_change
            let documentDirectory = try fileManager.url(for: .documentDirectory, in: .userDomainMask, appropriateFor:nil, create:false)
            let fileURL = documentDirectory.appendingPathComponent(filename)
            print("Saving to \(fileURL))")
            //let ctx = CGPDFContextCreateWithURL(fileURL, self.bounds, nil)
            __GraphicsBeginPDFContextToFile(fileURL.path, CGRect(x:0, y:0, width: rect.width, height: rect.height), nil)
            __GraphicsBeginPDFPage()
            let ctx = __GraphicsGetCurrentContext()!
            ctx.scaleBy(x: scale_change, y: scale_change)
            let rounded_origin = point(for: intPoint(at: CGPoint(x: bounds.minX, y: bounds.minY)))
            ctx.translateBy(x: -rounded_origin.x, y: -rounded_origin.y)
            //let savedMapOrigin = mapOrigin
            
            pdf_force_draw_grid = true
            draw(bounds)
            pdf_force_draw_grid = false
            
            __GraphicsEndPDFContext()
        } catch {
            Swift.print(error)
            return false
        }
        return false
    }

}


