//
//  ViewController.swift
//  Scary Pacman
//
//  Created by Nikao on 19/9/17.
//  Copyright © 2017 Algoseum. All rights reserved.
//

import UIKit

enum UserInfo: Int, CaseIterable {
    case leftStepper, rightStepper, endWithA0, lengths
}

class TriangularGridViewController: UIViewController {

    var myView : TriangularGridView { return view as! TriangularGridView }
    
    var simul: OpportunisticOritatamiSimulatorRemember // OritatamiSimulator
    
    var userInteractionEnabled = true {
        didSet {
            DispatchQueue.main.async { [weak self] in
                guard let me = self else { return }
                for b in [me.undoButton, me.clearButton] {
                    b?.isEnabled = me.userInteractionEnabled
                }
            }
        }
    }
    
    var rotateClockwise = 0
    let animated = false
    var allInputs = false
    static let follow = false
    var centerOrigin = false

//    @IBOutlet var filenameTableViewController : FilenameTableViewController!
//    var navigationFilenameTableViewController : UINavigationController? = nil
//    var popFilenameController: UIPopoverPresentationController? = nil
    
    required init?(coder aDecoder: NSCoder) {
        
        
        simul = OpportunisticOritatamiSimulatorRemember(width: TriangularGrid.Defaults.width, height: TriangularGrid.Defaults.height, oritatamiSystem: OritatamiSystem.CameleonInfiniteLineOS())
        
//        let je = JSONEncoder()
//        let data = try! je.encode(simul.oritatamiSystem)
//        print(String(data:data, encoding:.utf8)!)
        
//        let ct1 = """
//            [
//                [  "declare", "titi",
//                    ["declare", "toto",
//                        ["unfold", 0, 6, "A20"],
//                        [ "loop", 3, "A1", "A2", ]
//                    ],
//                    ["loop", 5, "toto"], "A3", ["loop", 7, "toto"]
//                ],
//                "titi", "toto", "titi",
//            ]
//        """
//        let ct2 = """
//            [["declare", "toto", "unfold", 0, 5, "A", "B", "C"], "toto", "B", "toto"]
//"""
//        let compactTranscriptJSON = "{\"value\": \(ct1)}"
//
//        let jsonData = ct1.data(using: .utf8)!
//        let nestedStringsAndInts = try! JSONDecoder().decode(NestedStringsAndInts.self, from: jsonData)
//        print(nestedStringsAndInts.explicitDescription)
//
//        let compactTranscript = try! JSONDecoder().decode(CompactTranscript.self, from: jsonData)
// // try? CompactTranscript(from: nestedStringsAndInts)
//        print(compactTranscript)
//        print(compactTranscript.expand())
//
//        exit(0)

        super.init(coder: aDecoder)

    }
    
    //.MARK: Reset simulation
    
    func resetSimul(os_json_encoding: String? = nil) {
        simulation_started = false
        simul.view = nil
        myView.grid = nil

        var os : OritatamiSystem
        
        if let os_encoding = (os_json_encoding ?? last_loaded_os_encoding) ?? nil {
            let OSencodedData = os_encoding.data(using: .utf8)!
            let OSjsonDecoder = JSONDecoder()
            
            os = try! OSjsonDecoder.decode(OritatamiSystem.self, from: OSencodedData)
            print("LENGTHS: \(os.periodCount) \(os.beadTypeIdsSequence.count) \(os.transcriptPrefixLength)")
        } else {
            os = OritatamiSystem.CameleonInfiniteLineOS()
        }
        
        show(info: "\(os.name) loaded", for: 2)
        
        // Rotating the seed
        for i in 0..<os.seedPath.count {
            os.seedPath[i] = os.seedPath[i].cw(rotateClockwise)
        }

        
        simul = OpportunisticOritatamiSimulatorRemember(
            width: TriangularGrid.Defaults.width,
            height: TriangularGrid.Defaults.height,
            oritatamiSystem: os)
        
        print("\nRULE:")
        print(simul.oritatamiSystem.rule.prettyDescription(with: simul.oritatamiSystem.beadTypeSet))
        
        print("\(simul.beadTypeSet.count) beads in total")
        print(simul.beadTypeSet)
        
        simul.noLoopOnBest = true
        
        myView.grid = simul.grid
        simul.view = myView
        
        let first = IntPoint.zero
        if centerOrigin {
            myView.centerOrigin()
        }
        simul.drawSeed(from: first)
        if centerOrigin {
            myView.center(at: simul.currentPosition.point)
        }
        myView.setNeedsDisplay()
        
        userInteractionEnabled = true
        
        simul.animated = animated
        simul.frameRate = Defaults.frameRate // 1/60
        simul.follow = TriangularGridViewController.follow
        
        updateButtons()
    }
    
    var saved_grid : TriangularGridProtocol? = nil
        
    override func viewDidLoad() {
        super.viewDidLoad()
        
        print("Document location:", documentURL())
        
        resetSimul()

        myView.controller = self
        myView.centerOrigin()

        tapRecognizer1 = UITapGestureRecognizer(target: self, action: #selector(tap1(_:)))
        tapRecognizer1.numberOfTapsRequired = 1
        myView.addGestureRecognizer(tapRecognizer1)
        
        tapRecognizer2 = UITapGestureRecognizer(target: self, action: #selector(tap2(_:)))
        tapRecognizer2.numberOfTapsRequired = 2
        myView.addGestureRecognizer(tapRecognizer2)

        tapRecognizer3 = UITapGestureRecognizer(target: self, action: #selector(tap3(_:)))
        tapRecognizer3.numberOfTapsRequired = 3
        myView.addGestureRecognizer(tapRecognizer3)
        
        let tapRecognizers = [ tapRecognizer1, tapRecognizer2, tapRecognizer3 ]
        for t in tapRecognizers {
            if let t = t {
                t.delegate = self
            }
        }
        
        tapRecognizerSaveView = UILongPressGestureRecognizer(target: self, action: #selector(tapSaveView(_:)))
        tapRecognizerSaveView.numberOfTouchesRequired = 1
//        myView.addGestureRecognizer(tapRecognizerSaveView)

        tapRecognizerSavePath = UILongPressGestureRecognizer(target: self, action: #selector(tapSavePath(_:)))
        tapRecognizerSavePath.numberOfTouchesRequired = 2
//        myView.addGestureRecognizer(tapRecognizerSavePath)

        panRecognizer = UIPanGestureRecognizer(target: self, action: #selector(pan(_:)))
        myView.addGestureRecognizer(panRecognizer)
        
        pinchRecognizer = UIPinchGestureRecognizer(target: self, action: #selector(pinch(_:)))
        myView.addGestureRecognizer(pinchRecognizer)
        
//        print("Wifi adress: \(getWifiAddress())")

//        var r = OritatamiRule(numberOfBeads: 12)
//        r.set(beadType: BeadType(c:0, d:.NE), attracts: BeadType(c:1, d:.NW))
//        print(r)

//        DispatchQueue.global(qos: .userInitiated).async { [weak self] in
//            print("Runnnnnn!!!!!")
//            self?.simul.nextStep()
//            print("Done!!!!!")
//        }

        prepareInfoLabel()
                        
        show(info: "Done")
        
        updateButtons()
        
        showHowTo()
    }
    
    @IBAction func showHowToForced(_ sender: Any) {
        showHowTo(forced: true)
        }
        
    var showHowToDone = false
    func showHowTo(defaultInfo info: String = "", forced: Bool = false) {
        if forced || !showHowToDone {
            print("Alert")
            showHowToDone = true
            let alert = UIAlertController(
                title: "How to load an Oritatami System",
                message: "Place your .\(os_file_extension) file in the OS Simulator folder in iOS Files app\n \n \n \n \n \n \n\n  or use AirDrop\n\n(this message will no longer be presented)",
                preferredStyle: .alert)
            
            alert.addAction(UIAlertAction(
                title: "Ok",
                style: .default,
                handler: nil))
            
            let filesIcon = __Image(named:"files-app-icon.png")
            let filesImageView = __ImageView(image: filesIcon)
            filesImageView.contentMode = .scaleAspectFit
            filesImageView.image = filesIcon
            filesImageView.translatesAutoresizingMaskIntoConstraints = false
            alert.view.addSubview(filesImageView)
            NSLayoutConstraint(item: filesImageView,
                               attribute: .centerX,
                               relatedBy: .equal,
                               toItem: alert.view,
                               attribute: .centerX,
                               multiplier: 1.0,
                               constant: 0.0).isActive = true
            NSLayoutConstraint(item: filesImageView, attribute: .width, relatedBy: .equal, toItem: nil, attribute: .notAnAttribute, multiplier: 0, constant: 50).isActive = true
            NSLayoutConstraint(item: filesImageView, attribute: .height, relatedBy: .equal, toItem: nil, attribute: .notAnAttribute, multiplier: 0, constant: 50).isActive = true
            NSLayoutConstraint(item: filesImageView,
                               attribute: .centerY,
                               relatedBy: .equal,
                               toItem: alert.view,
                               attribute: .centerY,
                               multiplier: 1.0,
                               constant: 0.0).isActive = true
            
            present(alert, animated: true)
        } else {
            show(info: info)
        }
    }
    
    // MARK: Load popover
    
    var selectedFile:URL? = nil
    
    @IBAction func showPopOverBox(_ sender: UIBarButtonItem) {
        guard !alreadyRunning else {
            show(info: Alerts.stopingSimulationFirst)
            simul.mustStop = true
            mustReset = true
            return
        }

        let os_files = get_os_files()
        
        if os_files.count == 0 {
            show(info: "No OS file in application folder :-(")
            return
        }
        
        let filenameNavigationViewController : UINavigationController = self.storyboard!.instantiateViewController(withIdentifier: "loadAnOSFileID") as! UINavigationController
        
        let closeButton = UIBarButtonItem(title: "Load", style: .done, target: self, action: #selector(dismissFilenameSelection))
        filenameNavigationViewController.topViewController?.navigationItem.rightBarButtonItem = closeButton
        
        let filenameController = filenameNavigationViewController.viewControllers[0] as! FilenameTableViewController
        filenameController.mother = self
        filenameController.filenames = os_files
        
        filenameNavigationViewController.modalPresentationStyle = .popover
        filenameNavigationViewController.preferredContentSize   = CGSize(width: 600, height: 600)
        
        let popoverPresentationViewController = filenameNavigationViewController.popoverPresentationController
        
        popoverPresentationViewController?.permittedArrowDirections = UIPopoverArrowDirection.any
        popoverPresentationViewController?.barButtonItem = sender
        
        present(filenameNavigationViewController, animated: true, completion: nil)
        
    }
    
    @IBAction func dismissFilenameSelection(_ sender: Any) {
        print("dismissing modal - selected file: \(selectedFile)")
        self.dismiss(animated: true, completion: { print("Dismissed modal") } )
        if let url = selectedFile {
            let _ = load(url)
        } else {
            show(info: "No OS file loaded")
        }
    }
    
    // MARK: RUN
        
    var mustReset = false
    var alreadyRunning = false
    var auto_flush_surrounded_beads = TriangularGridDictionary.Defaults.auto_flush_surrounded_beads
    func run() {
        // Test FIRST if seed is set up
        guard !alreadyRunning else { return }
        alreadyRunning = true
        DispatchQueue(label: "SimulQueue", qos: .userInitiated, attributes:
            DispatchQueue.Attributes()).async { [weak self] in
                var stillRunning = true
                while let me = self, me.play {
                    stillRunning = me.simul.nextStep()
                    if !stillRunning {
                        break
                    }
//                    if me.simul.currentIndex%50 == 0 {
//                        let progress = String(format: "%3.2f", 100 * Double(me.simul.currentIndex) / Double(me.simul.oritatamiSystem.beadSequenceCount))
//                        print("Current index: \(me.simul.currentIndex) [\(progress)%]")
//                    }
                    if me.auto_flush_surrounded_beads && me.simul.currentIndex%TriangularGridDictionary.Defaults.nb_steps_between_flushes_of_surrounded_beads == 0 {
                        me.simul.grid.flush_inaccessible_beads_in_all_dictionnaries()
                    }
                    if me.simul.currentIndex%1000 == 0 {
                        let progress = String(format: "%3.2f", 100 * Double(me.simul.currentIndex) / Double(me.simul.oritatamiSystem.beadSequenceCount))
                        DispatchQueue.main.async { [weak self] in
                            self?.show(info:"Current index: \(me.simul.currentIndex) [\(progress)%]")
                        }
                    }
                }
                var msg = ""
                if !stillRunning {
                    msg = "Done at index: \(self?.simul.currentIndex ?? -1)"
                    self?.play = false
                } else {
                    msg = "Pause at index: \(self?.simul.currentIndex ?? -1)"
                }
                self?.simul.flushPendingGif()
                print(msg)
                DispatchQueue.main.async { [weak self] in
                    self?.updatePlayPauseButton()
                    self?.show(info: msg)
                    if self?.mustReset ?? false {
                        self?.resetSimul()
                        self?.updateButtons()
                        self?.mustReset = false
                    }
                }
                self?.alreadyRunning = false
        }
    }
    
    override func didReceiveMemoryWarning() {
        super.didReceiveMemoryWarning()
        // Dispose of any resources that can be recreated.
    }

    func updateButtons() {
        updateFollowButton()
        updateAnimatedButton()
//        updateNameButton()
        updateLoopOnBestButton()
        updateGifRecordButton()
        updateShowViewButton()
        update_auto_flush_button()
    }
    
    @IBOutlet weak var auto_flush_button: UIBarButtonItem!
    @IBAction func flip_auto_flush(_ sender: UIBarButtonItem) {
        auto_flush_surrounded_beads = !auto_flush_surrounded_beads
        updateButtons()
        show(info: auto_flush_surrounded_beads ? "Auto-flush surrounded beads" : "Don't flush surrounded beads")
    }
    
    func update_auto_flush_button() {
        if #available(iOS 13.0, *) {
            auto_flush_button.image = __Image(systemName: auto_flush_surrounded_beads ? "trash.slash" : "trash.slash.fill")
        }
        // Fallback on earlier versions
    }


    
//    @IBOutlet weak var nameButton: UIBarButtonItem!
//    func updateNameButton() {
//        nameButton.title = simul.oritatamiSystem.name
//    }
   
    @IBOutlet weak var loopOnBestButton: UIBarButtonItem!
    @IBAction func setLoopOnBest(_ sender: Any) {
        simul.noLoopOnBest = !simul.noLoopOnBest
        updateLoopOnBestButton()
        show(info: "Loop on best configurations: \(simul.noLoopOnBest ? "Off" : "On")")
    }
    
    func updateLoopOnBestButton() {
        if #available(iOS 13.0, *) {
            loopOnBestButton.image = __Image(systemName: simul.noLoopOnBest ? "lightbulb" : "lightbulb.fill")
        } else {
            // Fallback on earlier versions
            show(info: simul.noLoopOnBest ? "Don't loop on best paths" : "Loop on best paths")
        }
    }
    
    @IBOutlet weak var centerButton: UIBarButtonItem!
    @IBAction func center(_ sender: UIBarButtonItem) {
        myView.centerIfNeeded(at: simul.currentPosition, animated: true)
        show(info: "Show nascent beads")
    }
    
    @IBOutlet var pinchRecognizer: UIPinchGestureRecognizer!
    
    private var pinchStartingScale: CGFloat = 1
    private var pinchStartingCenter = CGPoint.zero
    private var pinchStartingPosition = CGPoint.zero

    @IBAction func pinch(_ recognizer: UIPinchGestureRecognizer) {
        guard recognizer.numberOfTouches == 2 else { return }
        switch recognizer.state {
        case .began, .possible:
            pinchStartingScale = myView.scale
            pinchStartingCenter = myView.mapCenter
            pinchStartingPosition = pinchRecognizer.location(in: myView) - myView.bounds.center
        case .changed, .ended, .cancelled:
            let loc = pinchRecognizer.location(in: myView) - myView.bounds.center
            let scale = myView.scale
            let newScale = scale * recognizer.scale
            myView.scale = newScale
            myView.mapCenter = myView.mapCenter + pinchStartingPosition / scale - loc / newScale
//            myView.mapOrigin = loc + recognizer.scale * (myView.mapOrigin - pinchStartingPosition)
//            myView.mapCenter = myView.mapCenter - (loc - pinchStartingPosition) / myView.scale
            // Reset
            recognizer.scale = 1
            pinchStartingPosition = loc
        case .failed:
            myView.scale = pinchStartingScale
            myView.mapCenter = pinchStartingCenter
        default:
            break
        }
        showScale()
    }
    
    @IBOutlet weak var animatedButton: UIBarButtonItem!
    @IBAction func animatedOnOff(_ sender: Any) {
        guard userInteractionEnabled else { return }
        
        simul.animated = !simul.animated
        updateAnimatedButton()
        
        let info = "Animated: \(simul.animated ? "On" : "Off")"
        show(info: info)
    }
    func updateAnimatedButton() {
        if #available(iOS 13.0, *) {
            animatedButton.image = __Image(systemName: simul.animated ? "ant.fill" : "ant")
        } else {
            // Fallback on earlier versions
            animatedButton.title = simul.animated ? "Animated On" : "Animated Off"
        }
    }
    
//    @IBOutlet var tapRecognizer1: UITapGestureRecognizer!
//    @IBAction func tap1(_ recognizer: UITapGestureRecognizer) {
//        guard userInteractionEnabled else { return }
//
//        followOnOff()
//    }

    
    @IBOutlet weak var followButton: UIBarButtonItem!
    @IBAction func followOnOff() {
        guard userInteractionEnabled else { return }
        
        simul.follow = !simul.follow
        
        updateFollowButton()
        
        let info = "Follow: \(simul.follow ? "On" : "Off")"
        show(info: info)
    }

    func updateFollowButton() {
        if #available(iOS 13.0, *) {
            followButton.image = __Image(systemName: simul.follow ? "location.fill" : "location")
        } else {
            // Fallback on earlier versions
            followButton.title = simul.follow ? "Follow On" : "Follow Off"
        }
    }
    
    @IBOutlet var tapRecognizer1: UITapGestureRecognizer!
    @IBAction func tap1(_ recognizer: UITapGestureRecognizer) {
        show(info: "\(last_loaded_url?.lastPathComponent ?? "no filename")", for: 4)
    }

    
    @IBOutlet var tapRecognizer2: UITapGestureRecognizer!
    @IBAction func tap2(_ recognizer: UITapGestureRecognizer) {
        guard userInteractionEnabled else { return }
        
        let p = recognizer.location(in: myView)
        let q = myView.intPoint(at: p)
        if let b = simul.grid.beadTypeId(at: q) {
            show(info: "\(simul.beadTypeSet.beadType[b].name) at \(q)", for: 4)
        } else {
            show(info: "\(q)", for: 4)
        }
    }

    
    @IBOutlet var tapRecognizer3: UITapGestureRecognizer!
    @IBAction func tap3(_ recognizer: UITapGestureRecognizer) {
        guard userInteractionEnabled else { return }
        
        let p = recognizer.location(in: myView)
        let q = myView.intPoint(at: p)
        
        zoomInOut(zoomButton, centered: false)
        myView.center(at: q.point)
    }

    @IBOutlet weak var zoomButton: UIBarButtonItem!
    @IBAction func zoomInOut(_ sender: UIBarButtonItem, centered: Bool = true) {
        
        let center : IntPoint = myView.intPoint(at: CGPoint(x: myView.bounds.midX, y: myView.bounds.midY))
        
        let minScale = TriangularGridView.Defaults.minScale
        let midScale = TriangularGridView.Defaults.minDrawBeadScale
        let maxScale = TriangularGridView.Defaults.bondLength
        let scales = [minScale, midScale, maxScale]
        if myView.scale >= scales.last! {
            myView.scale = scales.first!
        } else {
            for scale in scales {
                if scale > myView.scale {
                    myView.scale = scale
                    break
                }
            }
        }
        if centered {
            myView.center(at: center.point)
        }
        showScale()
    }
    
    func showScale() {
        show(info:String(format:"Scale: %.2fx", myView.scale/10))
    }
    
    @IBOutlet var tapRecognizerSaveView: UILongPressGestureRecognizer!
    @IBAction func tapSaveView(_ recognizer: UILongPressGestureRecognizer) {
        saveView(recognizer)
    }

    @IBOutlet weak var saveViewButton: UIBarButtonItem!
    @IBAction func saveView(_ sender: Any) {
        let now = Date()
        let filename = "OSsimulatorView-\(String.with(date: now))"
        
        let pdfFilename = filename + ".pdf"
        let _ = myView.saveViewToPDFDocumentFile(pdfFilename)
        
        showSave(defaultInfo: "View saved in « \(filename) »")
    }
    
    
    @IBOutlet var tapRecognizerSavePath: UILongPressGestureRecognizer!
    @IBAction func tapSavePath(_ recognizer: UILongPressGestureRecognizer) {
        savePath(recognizer)
    }
    
    @IBOutlet weak var savePathButton: UIBarButtonItem!
    @IBAction func savePath(_ sender: Any) {
        let now = Date()
        let filename = "OSsimulatorPath-\(String.with(date: now))"
        
        let pdfFilename = filename + ".pdf"
        let _ = myView.saveToPathPDFDocumentFile(pdfFilename, compressedIfPossible: myView.scale <= TriangularGridView.Defaults.minDrawBeadScale)
        
        showSave(defaultInfo: "Path saved in « \(filename) »")
    }
    
    var showSavingDoneOnce = false
    func showSave(defaultInfo info: String = "") {
        if !showSavingDoneOnce {
            print("Alert")
            showSavingDoneOnce = true
            let alert = UIAlertController(
                title: "Image saved as PDF file",
                message: "in CAOS Simulator folder in iOS Files app\n \n \n \n \n \n (this message will no longer be presented)",
                preferredStyle: .alert)
            
            alert.addAction(UIAlertAction(
                title: "Ok",
                style: .default,
                handler: nil))
            
            let filesIcon = __Image(named:"files-app-icon.png")
            let filesImageView = __ImageView(image: filesIcon)
            filesImageView.contentMode = .scaleAspectFit
            filesImageView.image = filesIcon
            filesImageView.translatesAutoresizingMaskIntoConstraints = false
            alert.view.addSubview(filesImageView)
            NSLayoutConstraint(item: filesImageView,
                               attribute: .centerX,
                               relatedBy: .equal,
                               toItem: alert.view,
                               attribute: .centerX,
                               multiplier: 1.0,
                               constant: 0.0).isActive = true
            NSLayoutConstraint(item: filesImageView, attribute: .width, relatedBy: .equal, toItem: nil, attribute: .notAnAttribute, multiplier: 0, constant: 50).isActive = true
            NSLayoutConstraint(item: filesImageView, attribute: .height, relatedBy: .equal, toItem: nil, attribute: .notAnAttribute, multiplier: 0, constant: 50).isActive = true
            NSLayoutConstraint(item: filesImageView,
                               attribute: .centerY,
                               relatedBy: .equal,
                               toItem: alert.view,
                               attribute: .centerY,
                               multiplier: 1.0,
                               constant: -15.0).isActive = true
            
            present(alert, animated: true)
        } else {
            show(info: info)
        }
    }
    
    @IBOutlet weak var playButton: UIBarButtonItem!
    var play: Bool = false
    var simulation_started = false
    @IBAction func playPause(_ sender: Any) {
        play = !play
        updatePlayPauseButton()
        
        simulation_started = simulation_started || play
        
        simul.mustStop = !play
        if play && !alreadyRunning {
            run()
        }
        show(info: play ? "Play" : "Pause")
    }
    
    func updatePlayPauseButton() {
        if #available(iOS 13.0, *) {
            playButton.image = __Image(systemName: play ? "pause.fill" : "play.fill")
        } else {
            // Fallback on earlier versions
            playButton.image = nil
            playButton.title = play ? "||" : "▶︎"
        }
        if !show_view {
            self.show_hide_view(self)
        }
    }
    
    @IBOutlet weak var undoButton: UIBarButtonItem!
    @IBAction func undo(_ sender: UIBarButtonItem) {
        guard userInteractionEnabled else { return }
    }
    
    private var panStartingPosition = CGPoint.zero
    
    @IBOutlet var panRecognizer: UIPanGestureRecognizer!
    @IBAction func pan(_ recognizer: UIPanGestureRecognizer) {
        switch recognizer.state {
        case .began:
//            panStartingPosition = myView.mapOrigin
            panStartingPosition = myView.mapCenter
        case .changed, .possible, .ended, .cancelled:
//            myView.mapOrigin = panStartingPosition + recognizer.translation(in: myView)
            let t = recognizer.translation(in: myView) / myView.scale
            myView.mapCenter = panStartingPosition + CGPoint(x: -t.x, y: -t.y)
        case .failed:
//            myView.mapOrigin = panStartingPosition
            myView.mapCenter = panStartingPosition
        }
    }
    
    @IBOutlet weak var recordAsGifButton: UIBarButtonItem!
    
    func updateRecordAsGifButton() {
        recordAsGifButton.title = (recordAsGif ? "🎥" :"❌")
    }
    
    var recordAsGif : Bool = false { didSet { updateRecordAsGifButton() }}
    @IBAction func flipRecordAsGif(_ sender: UIBarButtonItem) {
        recordAsGif = !recordAsGif
    }
    
    
    @IBAction func play(_ sender: UIBarButtonItem) {
        guard userInteractionEnabled && simul.grid.pathStartingPoints.count > 0 else { return }
        userInteractionEnabled = false
        sender.isEnabled = false
    }
    
    func showFramerate() {
        show(info: String(format:"Frame rate: %.2f f/s", 1/simul.frameRate))
    }
    
    @IBAction func slower(_ sender: UIBarButtonItem) {
        simul.frameRate *= 1.25
        showFramerate()
    }
    
    @IBAction func faster(_ sender: UIBarButtonItem) {
        simul.frameRate /= 1.25
        showFramerate()
    }
    
    override func viewDidLayoutSubviews() {
        super.viewDidLayoutSubviews()
        myView.setNeedsDisplay()
    }
    
    // MARK: - Buttons
    
    func clear_configuration() {
        simulation_started = false
        myView.grid?.clear()
        myView.centerOrigin()
        simul.drawSeed(from: IntPoint.zero)
    }
    
    @IBOutlet weak var clearButton: UIBarButtonItem!
    @IBAction func clear(_ sender: Any, acceptAction: @escaping () -> Void = {}) {
        if !simulation_started {
            acceptAction()
            clear_configuration()
            return
        }
        
        let alert = UIAlertController(
            title: "Erase all?",
            message: "Please confirm",
            preferredStyle: .alert)
        
        alert.addAction(UIAlertAction(
            title: "Erase",
            style: .destructive,
            handler: { [weak self] action in
                acceptAction()
                if let self = self {
                    self.clear_configuration()
                }
        }))
        
        alert.addAction(UIAlertAction(
            title: "Cancel",
            style: .cancel,
            handler: nil))
        
        
        present(alert, animated: true)
    }
    
    @IBOutlet weak var infoLabel: UILabel!
    var infoTimer : Timer? = nil
    
    func show(info: String, for t: TimeInterval = Defaults.infoLabelShowTimeInterval) {
        infoLabel.text = "  \(info)  "
        showInfoLabel()
    }
    
    func showInfoLabel(for t: TimeInterval = Defaults.infoLabelShowTimeInterval) {
        if infoLabel.isHidden {
            CATransaction.setDisableActions(true)
            infoLabel.alpha = 0
            CATransaction.setDisableActions(false)
            infoLabel.isHidden = false
        }

        UIView.animate(withDuration: Defaults.infoLabelHideTransition) { [weak self] in
            self?.infoLabel.alpha = 1
        }
        
//        start(timer: &infoTimer) { [weak self] _ in self?.hideInfoLabel() } // iOS ≥ 10.0
        if let t = infoTimer, t.isValid {
            t.invalidate()
            infoTimer = nil
        }
        infoTimer = Timer.scheduledTimer(timeInterval: Defaults.infoLabelShowTimeInterval, target: self, selector: #selector(TriangularGridViewController.hideInfoLabel), userInfo: nil, repeats: false)
    }
    
    func stop(timer: inout Timer?) {
        if timer != nil && timer!.isValid {
            timer!.invalidate()
            timer = nil
        }
    }

//    func start(timer: inout Timer?, for t: TimeInterval = Defaults.infoLabelShowTimeInterval, block: @escaping (Timer) -> Void) {
//        stop(timer: &timer)
//        timer = Timer.scheduledTimer(withTimeInterval: t, repeats: false, block: block)
//    }

    @objc
    func hideInfoLabel() {
        stop(timer: &infoTimer)
        
        UIView.animate(withDuration: Defaults.infoLabelHideTransition, animations: { [weak self] in
            self?.infoLabel.alpha = 0
        }) { [weak self] (completed) in
            if completed {
                self?.infoLabel.isHidden = true
            }
        }
    }

    func prepareInfoLabel() {
        infoLabel.font = __Font.systemFont(ofSize: __Font.labelFontSize, weight: __Font.Weight.bold)
        infoLabel.textColor = Defaults.infoLabelTextColor
        infoLabel.layer.backgroundColor = Defaults.infoLabelBackgroundColor.cgColor
        infoLabel.layer.cornerRadius = __Font.labelFontSize*0.5
        infoLabel.layer.borderWidth = 0
        
        infoLabel.isHidden = true
        infoLabel.alpha = 0
    }
    
//    func checkInputTextField(_ field: UITextField) {
//        guard !alreadyRunning else {
//            show(info: Alerts.stopingSimulationFirst)
//            simul.mustStop = true
//            mustReset = true
//            return
//        }
//
//        let text = field.text!
//
//        guard text != "" else {
//            // init the text with current input
//            let seedString = input.reduce("") { $0 + "\($1),"}
//            field.text = String(seedString.dropLast())
//            return
//        }
//
//        let a = text.split(whereSeparator: {[",", ".", ";", " ", "\t", "-"].contains($0)})
//        input = a
//            .map() { Int($0) }
//            .filter() { $0 != nil }
//            .map() { $0! }
//
//        resetSimul()
//        updateButtons()
//    }

//    @IBOutlet weak var seedButton: UIBarButtonItem!
//    @IBAction func setSeed(_ sender: Any) {
//        guard !alreadyRunning else {
//            show(info: Alerts.stopingSimulationFirst)
//            simul.mustStop = true
//            mustReset = true
//            return
//        }
//
//        clic_red_pink_extreme_values = (clic_red_pink_extreme_values + 1)%OritatamiSystem.red_pink_extreme_values.count
//
//        updateSeedButton()
//        resetSimul()
//        return
//
//        print("Alert")
//        let alert = UIAlertController(
//            title: "Input values",
//            message: "Enter a list of numbers between 0 and \((1<<q)-1)",
//            preferredStyle: .alert)
//
//        print(input)
//        alert.addTextField(configurationHandler: checkInputTextField)
//        let seedTextField = alert.textFields!.first!
//        seedTextField.keyboardType = .decimalPad
//
//        alert.addAction(UIAlertAction(
//            title: "Ok",
//            style: .default,
//            handler: { [weak self] _ in self?.checkInputTextField(alert.textFields!.first!)}
//            )
//        )
//
//        present(alert, animated: true)
//    }
    
    
    // RotateCW
    
    @IBAction func rotateCW(_ sender: UIBarButtonItem) {
        guard !alreadyRunning else {
            show(info: Alerts.stopingSimulationFirst)
            simul.mustStop = true
            mustReset = true
            return
        }
        
        clear(sender) { [weak self] in
            guard let self = self else { return }
            self.rotateClockwise = (self.rotateClockwise + 1)%6
            for i in 0..<self.simul.oritatamiSystem.seedPath.count {
                self.simul.oritatamiSystem.seedPath[i] = self.simul.oritatamiSystem.seedPath[i].cw
            }
            self.show(info:"View rotated by: \(self.rotateClockwise*60)°")
        }
    }
    
    
    // Save & load OS
    
    func documentURL() -> URL {
        let fileManager = FileManager.default;
        let documents = fileManager.urls(for: .documentDirectory, in: .userDomainMask).first!
        
        return documents
    }
    
    func copyOSFilesFromBundleToDocuments() {
        if let resPath = Bundle.main.resourcePath {
            do {
                let dirContents = try FileManager.default.contentsOfDirectory(atPath: resPath)
                print(dirContents)
                let documentsURL = FileManager.default.urls(for: .documentDirectory, in: .userDomainMask).first!
                let osFiles = dirContents.filter { x in os_file_extensions.reduce(false) { $0 || x.hasSuffix("." + $1) } }
                for fileName in osFiles {
                    let sourceURL = Bundle.main.bundleURL.appendingPathComponent(fileName)
                    let destURL = documentsURL.appendingPathComponent(fileName)
                    do {
                        try FileManager.default.copyItem(at: sourceURL, to: destURL)
                    } catch { }
                }
            } catch { }
        }
    }
    
    var timeIntervalShowFilename = 2 * Defaults.infoLabelShowTimeInterval
    var currentLoadIndex: Int = -1
    enum FileError: Error { case error }
    var last_loaded_os_encoding : String? = nil
    
    @IBOutlet weak var reloadButton: UIBarButtonItem!
    @IBAction func reload(_ sender: Any) {
        guard !alreadyRunning else {
            show(info: Alerts.stopingSimulationFirst)
            simul.mustStop = true
            mustReset = true
            return
        }

        print(selectedFile)
        if let url = selectedFile, load(url) {
            return
        }
        showPopOverBox(reloadButton)
    }
    
    func load(_ url: URL) -> Bool {
        var os_encoding: String? = nil
        let filename = url.deletingPathExtension().lastPathComponent
        do {
            show(info: "Loading \(filename)", for: timeIntervalShowFilename)
            os_encoding = try String(contentsOf: url, encoding: .utf8)
            last_loaded_url = url
        } catch {
            show(info: "Loading \(filename) Failed!", for: timeIntervalShowFilename)
            return false
        }
        last_loaded_os_encoding = os_encoding
        
        guard !alreadyRunning else {
            show(info: Alerts.stopingSimulationFirst)
            simul.mustStop = true
            mustReset = true
            return false
        }
        
        resetSimul(os_json_encoding: last_loaded_os_encoding)
        return true
    }

    
    func get_os_files() -> [URL] {
        let fileManager = FileManager.default;
        let documents = fileManager.urls(for: .documentDirectory, in: .userDomainMask).first!
        
        // List all contents of directory and return as [String] OR nil if failed
        
        do {
            let files = try fileManager.contentsOfDirectory(at: documents, includingPropertiesForKeys: nil, options: .skipsHiddenFiles)
            var os_files = files.filter { os_file_extensions.contains($0.pathExtension) }
            os_files.sort(by: { $0.lastPathComponent <= $1.lastPathComponent })
            print(os_files)
            if os_files.count == 0 {
                copyOSFilesFromBundleToDocuments()
                let files = try fileManager.contentsOfDirectory(at: documents, includingPropertiesForKeys: nil, options: .skipsHiddenFiles)
                os_files = files.filter { $0.pathExtension == os_file_extension }
            }
            
            if os_files.count == 0 {
                DispatchQueue.main.async { [weak self] in
                    self?.show(info: "No OS file found in document folder", for: self?.timeIntervalShowFilename ?? 0)
                }
                throw FileError.error
            } else {
                return os_files
            }
        } catch {
            DispatchQueue.main.async { [weak self] in
                self?.show(info: "Listing OS files failed!", for: self?.timeIntervalShowFilename ?? 0)
            }
        }
        return []
    }
    
    var last_loaded_url : URL? = nil

    @IBAction func load_next(_ sender: UIBarButtonItem) {
        guard !alreadyRunning else {
            show(info: Alerts.stopingSimulationFirst)
            simul.mustStop = true
            mustReset = true
            return
        }
        
        let fileManager = FileManager.default;
        let documents = fileManager.urls(for: .documentDirectory, in: .userDomainMask).first!
        
        // List all contents of directory and return as [String] OR nil if failed
        
        do {
            let files = try fileManager.contentsOfDirectory(at: documents, includingPropertiesForKeys: nil, options: .skipsHiddenFiles)
            var os_files = files.filter { $0.pathExtension == os_file_extension }
            os_files.sort(by: { $0.lastPathComponent <= $1.lastPathComponent })
            print(os_files)
            if os_files.count == 0 {
                copyOSFilesFromBundleToDocuments()
                let files = try fileManager.contentsOfDirectory(at: documents, includingPropertiesForKeys: nil, options: .skipsHiddenFiles)
                os_files = files.filter { $0.pathExtension == os_file_extension }
            }
            
            if os_files.count == 0 {
                show(info: "No OS file found in document folder", for: timeIntervalShowFilename)
                throw FileError.error
            } else {
                currentLoadIndex = (currentLoadIndex+1)%os_files.count
                let url = os_files[currentLoadIndex]
                let _ = load(url)
            }
        } catch {
            show(info: "Loading failed!", for: timeIntervalShowFilename)
        }
    }

    
    @IBOutlet weak var gifRecordButton: UIBarButtonItem!
    
    func updateGifRecordButton() {
        if #available(iOS 13.0, *) {
            gifRecordButton.title = nil
            gifRecordButton.image = __Image(systemName: simul.saveAsGif ? "video.fill" : "video")
        } else {
            // Fallback on earlier versions
            gifRecordButton.title = simul.saveAsGif ? "Gif Rec" : "No Rec"
        }
    }
    @IBAction func gifRecordSwitch(_ sender: Any) {
        guard !alreadyRunning else {
            show(info: Alerts.stopSimulationFirst)
            return
        }

        simul.saveAsGif = !simul.saveAsGif
        updateGifRecordButton()
        show(info: simul.saveAsGif ? "Start recording steps as Gif" : "Stop recording steps as Gif")
    }
    
    @IBOutlet weak var showViewButton: UIBarButtonItem!
    var show_view : Bool = true
    @IBAction func show_hide_view(_ sender: Any) {
        show_view = !show_view
        myView.isHidden = !show_view
        updateButtons()
        myView.setNeedsDisplay()
        show(info: "Disable live view for faster simulation")
    }
    
    func updateShowViewButton() {
        if #available(iOS 13.0, *) {
            showViewButton.image = UIImage(systemName: show_view ? "eye" : "eye.slash")
        } else {
            // Fallback on earlier versions
            showViewButton.title =  show_view ? "Hide" : "Show"
        }
    }
//    @IBAction func save(_ sender: UIBarButtonItem) {
//        let formater = DateFormatter()
//        formater.dateFormat = "yyyy_MM_dd-HH_mm_ss_SS"
//        let now = Date()
//        let filename = formater.string(from: now)
//
//        show(info: "Saving \(filename)...", for: timeIntervalShowFilename)
//        saveConfiguration(to: filename)
//        show(info: "\(filename) saved", for: timeIntervalShowFilename)
//    }
//
//    func saveConfiguration(to filename: String) {
//        // let support = FileManager.default.urls(for: .applicationSupportDirectory, in: .userDomainMask).first!
//        guard let c = c else { return }
//
//        let documents = FileManager.default.urls(for: .documentDirectory, in: .userDomainMask).first!
//        let url = documents.appendingPathComponent(filename).appendingPathExtension("txt")
//        let s = c.stringEncoding()
//        try? s.write(to: url, atomically: true, encoding: .utf8)
//    }

    
    
}
    


//extension PacmanViewController {
//    // Return IP address of WiFi interface (en0) as a String, or `nil`
//    func getWifiAddress() -> String? {
//        var address : String?
//
//        // Get list of all interfaces on the local machine:
//        var ifaddr : UnsafeMutablePointer<ifaddrs>?
//        guard getifaddrs(&ifaddr) == 0 else { return nil }
//        guard let firstAddr = ifaddr else { return nil }
//
//        // For each interface ...
//        for ifptr in sequence(first: firstAddr, next: { $0.pointee.ifa_next }) {
//            let interface = ifptr.pointee
//
//            // Check for IPv4 or IPv6 interface:
//            let addrFamily = interface.ifa_addr.pointee.sa_family
//            if addrFamily == UInt8(AF_INET) || addrFamily == UInt8(AF_INET6) {
//
//                // Check interface name:
//                let name = String(cString: interface.ifa_name)
//                if  name == "en0" {
//
//                    // Convert interface address to a human readable string:
//                    var hostname = [CChar](repeating: 0, count: Int(NI_MAXHOST))
//                    getnameinfo(interface.ifa_addr, socklen_t(interface.ifa_addr.pointee.sa_len),
//                                &hostname, socklen_t(hostname.count),
//                                nil, socklen_t(0), NI_NUMERICHOST)
//                    address = String(cString: hostname)
//                }
//            }
//        }
//        freeifaddrs(ifaddr)
//
//        return address
//    }
//}
//
