//
//  UIImageFromView.swift
//  Scary Pacman
//
//  Created by Nikao on 1/6/18.
//  Copyright © 2018 Algoseum. All rights reserved.
//

import UIKit

import ImageIO
import MobileCoreServices

extension __Image {
    convenience init(view: __View) {
        
        __GraphicsBeginImageContext(view.frame.size)
        view.layer.render(in:__GraphicsGetCurrentContext()!)
        let image = __GraphicsGetImageFromCurrentImageContext()
        __GraphicsEndImageContext()
        self.init(cgImage: image!.cgImage!)
    }
}

typealias TimedImageSequence = [(__Image, CGFloat)]

extension TimedImageSequence {
    func saveAsGif(to url: URL, loopCount: Int) -> Bool {
        let gifDict = [kCGImagePropertyGIFLoopCount as String: loopCount]
        let fileProperties = [kCGImagePropertyGIFDictionary as String: gifDict]
        let cfURL = url as CFURL
        if let destination = CGImageDestinationCreateWithURL(cfURL, kUTTypeGIF, self.count, nil) {
            CGImageDestinationSetProperties(destination, fileProperties as CFDictionary?)
            for (image, delay) in self {
                let gifProperties = [kCGImagePropertyGIFDictionary as String: [kCGImagePropertyGIFDelayTime as String: delay]]
                CGImageDestinationAddImage(destination, image.cgImage!, gifProperties as CFDictionary?)
            }
            return CGImageDestinationFinalize(destination)
        }
        return false
    }

}

extension Array {
    mutating func shuffle(firstMElements m: Int) {
        let last = m >= count ? count-1 : m
        for i in 0..<last {
            let j = count - i
            let k = Int(arc4random())%j
            if k > 0 {
                let v = self[i+k]
                self[i+k] = self[i]
                self[i] = v
            }
        }
    }
}

//extension __Image {
//    static func saveAsGif(images: TimedImageSequence, to url: URL, loopCount: Int) -> Bool {
//        let gifDict = [kCGImagePropertyGIFLoopCount as String: loopCount]
//        let fileProperties = [kCGImagePropertyGIFDictionary as String: gifDict]
//        let cfURL = url as CFURL
//        if let destination = CGImageDestinationCreateWithURL(cfURL, kUTTypeGIF, images.count, nil) {
//            CGImageDestinationSetProperties(destination, fileProperties as CFDictionary?)
//            for (image, delay) in images {
//                let gifProperties = [kCGImagePropertyGIFDictionary as String: [kCGImagePropertyGIFDelayTime as String: delay]]
//                CGImageDestinationAddImage(destination, image.cgImage!, gifProperties as CFDictionary?)
//            }
//            return CGImageDestinationFinalize(destination)
//        }
//        return false
//    }
//}
