//
//  Utils.swift
//  Scary Pacman
//
//  Created by Nikao on 1/6/18.
//  Copyright © 2018 Algoseum. All rights reserved.
//

import UIKit

infix operator %% : MultiplicationPrecedence

extension Int {
    @inline(__always) static func %%(lhs: Int, rhs: Int) -> Int {
        // non-negative modulo
        if lhs >= 0 {
            return lhs % rhs
        } else {
            return (lhs % rhs) + rhs
        }
    }
    
    func log2() -> Int {
        for i in 0...8*MemoryLayout<Int>.size {
            if self>>i <= 0 { return i }
        }
        return -1
    }
}

struct FastQueue<T> : CustomStringConvertible {
    var head: Int = 0
    var values: [T] = []
    var empty: Bool { return head >= values.count }
    
    mutating func push(_ x: T) {
        values.append(x)
    }
    
    mutating func pop() -> T? {
        guard !empty else { return nil }
        let x = values[head]
        head += 1
        return x
    }
    
    var description: String {
        return values[head..<values.count].description
    }
}

extension String {
    //
    static func with(date: Date) -> String {
        let dateFormatter = DateFormatter()
        dateFormatter.dateFormat = "yyyy_MM_dd-HH_mm_ss"
        
        return dateFormatter.string(from: date)
    }
    
    func timeStamp() -> String {
        let s = self + "-" + String.with(date: Date())
        print(s)
        return s
    }
    
    // padding
    func rightJustified(width: Int, truncate: Bool = false) -> String {
        guard width > count else {
            return truncate ? String(suffix(width)) : self
        }
        return String(repeating: " ", count: width - count) + self
    }
    
    func leftJustified(width: Int, truncate: Bool = false) -> String {
        guard width > count else {
            return truncate ? String(prefix(width)) : self
        }
        return self + String(repeating: " ", count: width - count)
    }

    func centered(width: Int, truncate: Bool = false) -> String {
        let diff = width - count
        let diff2 = diff/2
        guard diff >= 0 else {
            return truncate ? String(String(prefix(width-diff2)).suffix(width)) : self
        }
        return String(repeating: " ", count: diff-diff2) + self + String(repeating: " ", count: diff2)
    }
}

extension URL {
    static func documentURL(for filename: String) -> URL? {
        let fileManager = FileManager.default
        do {
            let documentDirectory = try fileManager.url(for: .documentDirectory, in: .userDomainMask, appropriateFor:nil, create:false)
            let url = documentDirectory.appendingPathComponent(filename)
//            print("URL:\n\t\(url)\n\tdocumentDir: \(documentDirectory)\n\tfilename: \(filename)")
            return url
        } catch {
            Swift.print(error)
            return nil
        }
    }
}
